"""
The object definitions for reading and writing data in single conceptual units
using an interface based on slicing definitions and numpy arrays with formatting
operations.

This module introduced in version 1.3.0.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import os
from typing import Union, Tuple, Sequence, BinaryIO, Optional

import numpy

from sarpy.io.general.format_function import FormatFunction, IdentityFunction
from sarpy.io.general.slice_parsing import verify_subscript, get_slice_result_size, \
    get_subscript_result_size
from sarpy.io.general.utils import h5py, is_file_like

if h5py is not None:
    from h5py import File as h5pyFile, Dataset as h5pyDataset
else:
    h5pyFile = None
    h5pyDataset = None

logger = logging.getLogger(__name__)


####
# helper functions

def _reverse_slice(slice_in: slice) -> slice:
    """
    Given a slice with negative step, this returns a slice which will define the
    same elements traversed in the opposite direction. Note that this is not
    the same as the mirror operation.

    Parameters
    ----------
    slice_in : slice

    Returns
    -------
    slice

    Raises
    ------
    ValueError
    """

    if slice_in.step > 0:
        raise ValueError('This is only applicable to slices with negative step value')
    stop = 0 if slice_in.stop is None else slice_in.stop + 1
    mult = int(numpy.floor((stop - slice_in.start)/slice_in.step))
    final_entry = slice_in.start + mult*slice_in.step
    return slice(final_entry, slice_in.start-slice_in.step, -slice_in.step)


def _find_slice_overlap(
        slice_in: slice,
        ref_slice: slice) -> Tuple[Optional[slice], Optional[slice]]:
    """
    Finds the overlap of the slice with a contiguous interval slice.

    Parameters
    ----------
    slice_in : slice
    ref_slice : slice

    Returns
    -------
    child_slice : None|slice
        The overlap expressed as a slice relative to the sliced coordinates.
        None if there is no overlap.
    parent_slice : None|slice
        The overlap expressed as a slice relative to the overall indices.
        None if there is no overlap.
    """

    if ref_slice.step not in [1, -1]:
        raise ValueError('Reference slice must have step +/-1')

    if ref_slice.step > 0:
        start_ind = ref_slice.start
        stop_ind = ref_slice.stop
    else:
        start_ind = 0 if ref_slice.stop is None else ref_slice.stop + 1
        stop_ind = ref_slice.start + 1

    if slice_in.step > 0:
        if slice_in.stop <= start_ind or slice_in.start >= stop_ind:
            # there is no overlap

            return None, None
        # find minimum multiplier so that slice_in.start + mult*slice_in.step >= start_ind
        #    mult >= (start_ind - slice_in.start)/slice_in.step
        child_start = 0 if start_ind <= slice_in.start else \
            int(numpy.ceil((start_ind - slice_in.start)/slice_in.step))
        parent_start = slice_in.start - start_ind + child_start*slice_in.step
        # find end - maximum multiplier so that slice_in.start + mult*slice_in.step <= stop_ind
        #    mult <= (stop_ind - slice_in.start)/slice_in.step
        max_ind = min(slice_in.stop, stop_ind)
        child_stop = int(numpy.floor((max_ind - slice_in.start)/slice_in.step))
        parent_stop = slice_in.start - start_ind + child_stop*slice_in.step
    else:
        if slice_in.start < start_ind or (slice_in.stop is not None and slice_in.stop >= stop_ind):
            # there is no overlap

            # noinspection PyTypeChecker
            return None, None

        # find minimum multiplier so that slice_in.start + mult*slice_in.step <= stop_ind-1
        #    mult >= (stop_ind - 1 - slice_in.start)/slice_in.step
        child_start = 0 if slice_in.start < stop_ind else int(numpy.ceil((stop_ind - 1 - slice_in.start)/slice_in.step))
        parent_start = slice_in.start - start_ind + child_start*slice_in.step
        # find end - first multiplier so that slice_in.start + mult*slice_in.step < start_ind - 1
        #    mult > (start_ind - slice_in.start)/slice_in.step
        if slice_in.stop is None:
            min_ind = max(start_ind-1, -1)
        else:
            min_ind = max(start_ind-1, slice_in.stop)
        child_stop = int(numpy.ceil((min_ind - slice_in.start)/slice_in.step))
        parent_stop = slice_in.start - start_ind + child_stop*slice_in.step
        if parent_stop < 0:
            parent_stop = None

    if ref_slice.step < 0:
        # noinspection PyTypeChecker
        return _reverse_slice(slice(parent_start, parent_stop, slice_in.step)), \
               _reverse_slice(slice(child_start, child_stop, 1))
    else:
        # noinspection PyTypeChecker
        return slice(parent_start, parent_stop, slice_in.step), \
               slice(child_start, child_stop, 1)


def _infer_subscript_for_write(
        data: numpy.ndarray,
        start_indices: Union[None, int, Tuple[int, ...]],
        subscript: Union[None, Sequence[slice]],
        full_shape: Tuple[int, ...]) -> Tuple[slice, ...]:
    """
    Helper function, for writing operation, which infers the subscript definition
    between the given start_indices or (possibly partially defined) subscript.

    Parameters
    ----------
    data : numpy.ndarray
    start_indices : None|int|Tuple[int, ...]
    subscript : None|Sequence[slice]
    full_shape : Tuple[int, ...]

    Returns
    -------
    Tuple[slice, ...]
    """

    if start_indices is None and subscript is None:
        if data.shape == full_shape:
            return verify_subscript(None, full_shape)
        else:
            raise ValueError('One of start_indices or subscript must be provided.')

    if start_indices is not None:
        if isinstance(start_indices, int):
            start_indices = (start_indices, )

        if len(start_indices) < len(full_shape):
            start_indices = start_indices + tuple(0 for _ in range(len(start_indices), len(full_shape)))
        subscript = tuple([slice(entry1, entry1+entry2, 1) for entry1, entry2 in zip(start_indices, data.shape)])
    subscript, result_shape = get_subscript_result_size(subscript, full_shape)
    if result_shape != data.shape:
        raise ValueError(
            'Inferred subscript `{}` with shape `{}`\n\t'
            'does not match data.shape `{}`'.format(subscript, result_shape, data.shape))
    return subscript


def extract_string_from_subscript(
        subscript: Union[None, int, slice, Tuple]) -> Tuple[Union[None, int, slice, Sequence], Tuple[str, ...]]:
    """
    Extracts any string elements (stripped and made all lowercase) from subscript entries.

    Parameters
    ----------
    subscript : None|str|int|slice|Sequence

    Returns
    -------
    subscript: None|int|slice|Sequence
        With string entries removed
    strings : Tuple[str, ...]
        The string entries, stripped and made all lower case.
    """

    string_entries = []
    if isinstance(subscript, str):
        string_entries.append(subscript.strip().lower())
        subscript = None
    elif isinstance(subscript, Sequence):
        new_subscript = []
        for entry in subscript:
            if isinstance(entry, str):
                string_entries.append(entry.strip().lower())
            else:
                new_subscript.append(entry)
        if len(string_entries) > 0:
            subscript = tuple(new_subscript)
    return subscript, tuple(string_entries)


#####
# Abstract data segment definition and derived element implementations

class DataSegment(object):
    """
    Partially abstract base class representing one conceptual fragment of data
    read or written as an array. This is generally designed for images, but is
    general enough to support other usage.

    Introduced in version 1.3.0.

    .. warning::
        The format function instance will be modified in place. Do not use the
        same format function instance across multiple data segments.
    """

    _allowed_modes = ('r', 'w')

    __slots__ = (
        '_closed', '_mode',
        '_raw_dtype', '_raw_shape', '_formatted_dtype', '_formatted_shape',
        '_reverse_axes', '_transpose_axes', '_reverse_transpose_axes',
        '_format_function')

    def __init__(
            self,
            raw_dtype: Union[str, numpy.dtype],
            raw_shape: Tuple[int, ...],
            formatted_dtype: Union[str, numpy.dtype],
            formatted_shape: Tuple[int, ...],
            reverse_axes: Union[None, int, Sequence[int]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            mode: str = 'r'):
        """

        Parameters
        ----------
        raw_dtype : str|numpy.dtype
        raw_shape : Tuple[int, ...]
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform to the raw data, after applying
            any axis reversal, and before applying any format function
        format_function : None|FormatFunction
            Note that the format function instance
        mode : str
        """

        self._closed = False
        self._mode = None
        self._set_mode(mode)

        self._raw_shape = None
        self._set_raw_shape(raw_shape)

        self._raw_dtype = None
        self._set_raw_dtype(raw_dtype)

        self._formatted_shape = None
        self._set_formatted_shape(formatted_shape)

        self._formatted_dtype = None
        self._set_formatted_dtype(formatted_dtype)

        self._reverse_axes = None
        self._set_reverse_axes(reverse_axes)

        self._transpose_axes = None
        self._reverse_transpose_axes = None
        self._set_transpose_axes(transpose_axes)

        self._format_function = None
        self._set_format_function(format_function)
        self._validate_shapes()

    @property
    def raw_shape(self) -> Tuple[int, ...]:
        """
        Tuple[int, ...]: The raw shape.
        """

        return self._raw_shape

    def _set_raw_shape(self, value: Tuple[int, ...]) -> None:
        if not isinstance(value, tuple):
            raise TypeError(
                'raw_shape must be specified by a tuple of ints, got type `{}`'.format(type(value)))
        for entry in value:
            if not isinstance(entry, int):
                raise TypeError(
                    'raw_shape must be specified by a tuple of ints, got `{}`'.format(value))
            if entry <= 0:
                raise ValueError(
                    'raw_shape must be specified by a tuple of positive ints, got `{}`'.format(value))
        self._raw_shape = value

    @property
    def raw_ndim(self) -> int:
        """
        int: The number of raw dimensions.
        """

        return len(self._raw_shape)

    @property
    def mode(self) -> str:
        """
        str: The mode.
        """

        return self._mode

    def _set_mode(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError('Mode must be a string value')
        value = value.strip().lower()

        if value not in self._allowed_modes:
            raise ValueError('mode must be one of {}'.format(self._allowed_modes))

        self._mode = value

    @property
    def raw_dtype(self) -> numpy.dtype:
        """
        numpy.dtype: The data type of the data returned by the :func:`read_raw` function.
        """

        return self._raw_dtype

    def _set_raw_dtype(self, value) -> None:
        if not isinstance(value, numpy.dtype):
            try:
                value = numpy.dtype(value)
            except Exception as e:
                raise ValueError(
                    'Tried interpreting raw_dtype value as a numpy.dtype, '
                    'and failed with error\n\t{}'.format(e))
        self._raw_dtype = value

    @property
    def formatted_shape(self) -> Tuple[int, ...]:
        """
        Tuple[int, ...]: The formatted data shape.
        """

        return self._formatted_shape

    def _set_formatted_shape(self, value: Tuple[int, ...]) -> None:
        if not isinstance(value, tuple):
            raise TypeError(
                'formatted_shape must be specified by a tuple of ints, got type `{}`'.format(type(value)))
        for entry in value:
            if not isinstance(entry, int):
                raise TypeError(
                    'formatted_shape must be specified by a tuple of ints, got `{}`'.format(value))
            if entry <= 0:
                raise ValueError(
                    'formatted_shape must be specified by a tuple of positive ints, got `{}`'.format(value))
        self._formatted_shape = value

    @property
    def formatted_dtype(self) -> numpy.dtype:
        """
        numpy.dtype: The data type of the formatted data, which will be returned
        by the :func:`read` function.
        """

        return self._formatted_dtype

    def _set_formatted_dtype(self, value) -> None:
        if not isinstance(value, numpy.dtype):
            try:
                value = numpy.dtype(value)
            except Exception as e:
                raise ValueError(
                    'Tried interpreting formatted_dtype value as a numpy.dtype, '
                    'and failed with error\n\t{}'.format(e))
        self._formatted_dtype = value

    @property
    def formatted_ndim(self) -> int:
        """
        int: The number of formatted dimensions.
        """

        return len(self._formatted_shape)

    @property
    def reverse_axes(self) -> Optional[Tuple[int, ...]]:
        """
        None|Tuple[int, ...]: The collection of axes (with respect to raw order)
        along which we will reverse as part of transformation to formatted data order.
        If not `None`, then this will be a tuple in strictly increasing order.
        """

        return self._reverse_axes

    def _set_reverse_axes(self, value: Union[None, int, Tuple[int, ...]]) -> None:
        if value is None:
            self._reverse_axes = None
            return

        if isinstance(value, int):
            value = (value, )
        else:
            value = tuple(sorted(list(set(int(entry) for entry in value))))

        for entry in value:
            if not (0 <= entry < self.raw_ndim):
                raise ValueError('reverse_axes entries must be less than raw_ndim')

        self._reverse_axes = value

    @property
    def transpose_axes(self) -> Tuple[int, ...]:
        """
        None|Tuple[int, ...]: The transpose order for switching from raw order to
        formatted order, prior to applying any format function.

        If populated, this must be a permutation of `(0, 1, ..., raw_ndim-1)`.
        """

        return self._transpose_axes

    def _set_transpose_axes(self, value: Union[None, Tuple[int, ...]]) -> None:
        if value is None:
            self._transpose_axes = None
            return
        value = tuple([int(entry) for entry in value])
        if set(value) != set(range(self.raw_ndim)):
            raise ValueError('transpose_axes must be a permutation of range(raw_ndim), got\n\t{}'.format(value))
        self._transpose_axes = value

    @property
    def format_function(self) -> FormatFunction:
        """
        The format function which will be applied to the raw data.

        Returns
        -------
        FormatFunction
        """

        return self._format_function

    def _set_format_function(self, value: Optional[FormatFunction]) -> None:
        if value is None:
            value = IdentityFunction()
        if not isinstance(value, FormatFunction):
            raise ValueError('Got unexpected format_function value of type `{}`'.format(type(value)))

        # set our important property values
        value.set_raw_shape(self.raw_shape)
        value.set_formatted_shape(self.formatted_shape)
        value.set_reverse_axes(self.reverse_axes)
        value.set_transpose_axes(self.transpose_axes)
        self._format_function = value

    @property
    def can_write_regular(self) -> bool:
        """
        bool: Can this data segment write regular data, which requires a function
        inverse?
        """

        return self.mode == 'w' and self.format_function.has_inverse

    @property
    def closed(self) -> bool:
        """
        bool: Is the data segment closed? Reading or writing will result in a ValueError
        """

        return self._closed

    def _validate_closed(self):
        if not hasattr(self, '_closed') or self._closed:
            raise ValueError('I/O operation of closed data segment')

    def _validate_shapes(self) -> None:
        """
        Validate the raw_shape and formatted_shape values.
        """

        self.format_function.validate_shapes()

    # read related methods
    def verify_raw_subscript(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]]) -> Tuple[slice, ...]:
        """
        Verifies that the structure of the subscript is in keeping with the raw
        shape, and fills in any missing dimensions.

        Parameters
        ----------
        subscript : None|int|slice|Sequence[int|slice|Tuple[int, ...]]

        Returns
        -------
        Tuple[slice, ...]
            Guaranteed to be a tuple of slices of length `raw_ndim`.
        """

        return verify_subscript(subscript, self._raw_shape)

    def verify_formatted_subscript(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]]) -> Tuple[slice, ...]:
        """
        Verifies that the structure of the subscript is in keeping with the formatted
        shape, and fills in any missing dimensions.

        Parameters
        ----------
        subscript : None|int|slice|Sequence[int|slice|Tuple[int, ...]]

        Returns
        -------
        Tuple[slice, ...]
            Guaranteed to be a tuple of slices of length `formatted_ndim`.
        """

        return verify_subscript(subscript, self._formatted_shape)

    def _interpret_subscript(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            raw: bool = False) -> Tuple[slice, ...]:
        """
        Restructures the subscript to be a tuple of slices guaranteed to be the same
        length as the dimension of the return.

        Parameters
        ----------
        subscript : None|int|slice|Tuple[slice, ...]
        raw : bool
            If `True` then this should apply to raw coordinates, if `False` it
            should apply to original coordinates.

        Returns
        -------
        Tuple[slice, ...]
        """

        if raw:
            return verify_subscript(subscript, self._raw_shape)
        else:
            return verify_subscript(subscript, self._formatted_shape)

    def __getitem__(
            self,
            subscript: Union[None, int, slice, str, Sequence[Union[None, int, slice, str]]]) -> numpy.ndarray:
        """
        Fetch the data via slice definition.

        Parameters
        ----------
        subscript : None|int|slice|str|tuple

        Returns
        -------
        numpy.ndarray
        """

        # TODO: document the string entries situation
        self._validate_closed()

        subscript, string_entries = extract_string_from_subscript(subscript)
        use_raw = ('raw' in string_entries)
        squeeze = ('nosqueeze' not in string_entries)

        if use_raw:
            return self.read_raw(subscript, squeeze=squeeze)
        else:
            return self.read(subscript, squeeze=squeeze)

    def read(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        """
        In keeping with data segment mode, read the data slice specified relative
        to the formatted data coordinates. This requires that `mode` is `'r'`.

        Parameters
        ----------
        subscript : None|int|slice|Sequence[int|slice|Tuple[int, ...]]
        squeeze : bool
            Apply the numpy.squeeze operation, which eliminates dimension of size 1?

        Returns
        -------
        numpy.ndarray
        """

        self._validate_closed()

        if self.mode != 'r':
            raise ValueError('Requires mode = "r"')
        norm_subscript = self.verify_formatted_subscript(subscript)
        raw_subscript = self.format_function.transform_formatted_slice(norm_subscript)
        raw_data = self.read_raw(raw_subscript, squeeze=False)
        return self.format_function(raw_data, raw_subscript, squeeze=squeeze)

    # noinspection PyTypeChecker
    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        """
        In keeping with data segment mode, read raw data from the source, without
        reformatting and or applying symmetry operations. This requires that `mode`
        is `'r'`.

        Parameters
        ----------
        subscript : None|int|slice|Sequence[int|slice|Tuple[int, ...]]
            These arguments are relative to raw data shape and order, no symmetry
            operations have been applied.
        squeeze : bool
            Apply numpy.squeeze, which eliminates any dimensions of size 1?

        Returns
        -------
        numpy.ndarray
            This will be of data type given by `raw_dtype`.
        """

        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        raise NotImplementedError

    def _verify_write_raw_details(self, data: numpy.ndarray) -> None:
        if self.mode != 'w':
            raise ValueError('I/O Error, functionality requires mode == "w"')

        if data.dtype.itemsize != self.raw_dtype.itemsize:
            raise ValueError(
                'Expected data dtype itemsize `{}`, got `{}`'.format(self.raw_dtype.itemsize, data.dtype.itemsize))
        if data.dtype != self.raw_dtype:
            logger.warning('Expected data dtype `{}`, got `{}`.'.format(self.raw_dtype, data.dtype))

    def write(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs) -> None:
        """
        In keeping with data segment mode, write the data provided in formatted
        form, assuming the slice specified relative to the formatted data coordinates.

        This requires that `mode` is `'w'`, and `format_function.has_inverse == True`,
        because we have to apply the format function inverse to the provided data.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data in formatted form, to be transferred to raw form and written.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Sequence[slice]
            The subscript definition in formatted coordinates.
        kwargs

        Returns
        -------
        None
        """

        self._validate_closed()

        if not self.can_write_regular:
            raise ValueError(
                'I/O error, functionality requires mode = "w"\n\t'
                'and the ability to invert the format function')

        if data.dtype.itemsize != self.formatted_dtype.itemsize:
            raise ValueError(
                'Expected data dtype itemsize `{}`, got `{}`'.format(
                    self.formatted_dtype.itemsize, data.dtype.itemsize))
        if data.dtype != self.formatted_dtype:
            logger.warning('Expected data dtype `{}`, got `{}`.'.format(
                self.formatted_dtype, data.dtype))

        subscript = _infer_subscript_for_write(
            data, start_indices, subscript, self.formatted_shape)

        raw_data = self.format_function.inverse(data, subscript)
        raw_subscript = self.format_function.transform_formatted_slice(subscript)
        self.write_raw(raw_data, subscript=raw_subscript, **kwargs)

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs) -> None:
        """
        In keeping with data segment mode, write the data provided in raw form,
        assuming the slice specified relative to raw data coordinates. This
        requires that `mode` is `'w'`.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data in raw form.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Tuple[slice, ...]
            The subscript definition in raw coordinates.
        kwargs

        Returns
        -------
        None
        """

        raise NotImplementedError

    def check_fully_written(self, warn: bool = False) -> bool:
        """
        Checks that all expected pixel data is fully written.

        Parameters
        ----------
        warn : bool
            Log warning with some details if not fully written.

        Returns
        -------
        bool
        """

        raise NotImplementedError

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        """
        This returns the bytes for the underlying raw data.

        .. warning::
            A data segment is *conceptually* represented in raw data as a single
            numpy array of appropriate shape and data type. When the data segment
            is formed from component pieces, then the return of this function may
            deviate significantly from the raw byte representation of such an
            array after consideration of data order and pad pixels.

        Parameters
        ----------
        warn : bool
            If `True`, then a check will be performed to ensure that the data
            has been fully written and warnings printed if the answer is no.

        Returns
        -------
        bytes|Tuple
            The result will be a `bytes` object, unless the data segment is
            made up of a collection of child data segments, in which case the
            result will be a Tuple consisting of their `get_raw_bytes` returns.
        """

        raise NotImplementedError

    def flush(self) -> None:
        """
        Should perform, if possible, any necessary steps to flush any unwritten
        data to the file.

        Returns
        -------
        None
        """

        return

    def close(self):
        """
        This should perform any necessary clean-up operations, like closing
        open file handles, deleting any temp files, etc
        """

        if not hasattr(self, '_closed') or self._closed:
            return
        self._closed = True

    def __del__(self):
        # NB: this is called when the object is marked for garbage collection
        # (i.e. reference count goes to 0)
        # This order in which this happens may be unreliable
        self.close()


class ReorientationSegment(DataSegment):
    """
    Define a basic ordering of a given DataSegment. The raw data will be
    presented as the parent data segment's formatted data.

    Introduced in version 1.3.0.
    """

    __slots__ = ('_parent', '_close_parent')

    def __init__(
            self,
            parent: DataSegment,
            formatted_dtype: Optional[Union[str, numpy.dtype]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            close_parent: bool = True):
        """
        Parameters
        ----------
        parent : DataSegment
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform, after applying any axis reversal,
            and before applying any format function
        close_parent : bool
            Call parent.close() when close is called?
        """

        self._parent = None
        self._close_parent = None
        self.close_parent = close_parent
        intermediate_shape = self._set_parent(parent, transpose_axes)
        if format_function is None:
            formatted_dtype = parent.formatted_dtype
            formatted_shape = intermediate_shape
        else:
            if formatted_dtype is None or formatted_shape is None:
                raise ValueError(
                    'If format_function is provided,\n\t'
                    'then formatted_dtype and formatted_shape must be provided.')
        mode = parent.mode
        if mode == 'w' and not parent.can_write_regular:
            raise ValueError('Requires that the parent can write regular data')

        DataSegment.__init__(
            self, parent.formatted_dtype, parent.formatted_shape, formatted_dtype, formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode=mode)

    @property
    def parent(self) -> DataSegment:
        return self._parent

    def _set_parent(self,
                    parent: DataSegment,
                    transpose_axes: Union[None, Tuple[int, ...]]) -> Tuple[int, ...]:
        if transpose_axes is None:
            trans_axes = tuple(range(parent.formatted_ndim))
        else:
            if len(transpose_axes) != parent.formatted_ndim:
                raise ValueError('transpose_axes must have length {}'.format(parent.formatted_ndim))
            trans_axes = transpose_axes
        self._parent = parent
        return tuple([parent.formatted_shape[index] for index in trans_axes])

    @property
    def close_parent(self) -> bool:
        """
        bool: Call parent.close() when close is called?
        """

        return self._close_parent

    @close_parent.setter
    def close_parent(self, value):
        self._close_parent = bool(value)

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:

        self._validate_closed()
        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        return self.parent.read(subscript, squeeze=squeeze)

    def check_fully_written(self, warn: bool = False) -> bool:
        return self.parent.check_fully_written(warn=warn)

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs):
        """
        In keeping with data segment mode, write the data provided in raw form,
        assuming the slice specified relative to raw data coordinates. This
        requires that `mode` is `'w'`.

        Note that raw form/order for the data segment is simply a reordered version
        of the formatted data for parent. This **is not** related to raw data
        with respect to the parent. To write raw data with respect to the parent,
        use :func:`parent.write_raw` instead.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data in raw form.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Sequence[slice]
            The subscript definition in raw coordinates.
        kwargs

        Returns
        -------
        None
        """

        self._validate_closed()
        self._verify_write_raw_details(data)

        subscript = _infer_subscript_for_write(data, start_indices, subscript, self.raw_shape)
        parent_form_subscript = self.format_function.transform_formatted_slice(subscript)
        self.parent.write(data, subscript=parent_form_subscript, **kwargs)

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        self._validate_closed()
        return self.parent.get_raw_bytes(warn=warn)

    def flush(self) -> None:
        self._validate_closed()
        try:
            self.parent.flush()
        except AttributeError:
            return

    def close(self):
        try:
            if self._closed:
                return

            self.flush()
            if self.close_parent:
                self.parent.close()
            DataSegment.close(self)
            self._parent = None
        except AttributeError:
            return


class SubsetSegment(DataSegment):
    """
    Define a subset of a given DataSegment, with formatting handled by the
    parent data segment.

    Introduced in version 1.3.0.
    """
    _allowed_modes = ('r', 'w')

    __slots__ = (
        '_parent', '_formatted_subset_definition', '_raw_subset_definition',
        '_original_formatted_indices', '_original_raw_indices', '_squeeze',
        '_close_parent', '_pixels_written', '_expected_pixels_written')

    def __init__(
            self,
            parent: DataSegment,
            subset_definition: Tuple[slice, ...],
            coordinate_basis: str,
            squeeze: bool = True,
            close_parent: bool = True):
        """
        Parameters
        ----------
        parent : DataSegment
        subset_definition : Tuple[slice, ...]
        coordinate_basis : str
            The coordinate basis for the subset definition, it should be one of
            `('raw', 'formatted')`.
        squeeze: bool
            Eliminate the dimensions that are size 1 in the subset?
        close_parent : bool
            Call parent.close() when close is called?
        """

        self._close_parent = None
        self.close_parent = close_parent
        self._original_formatted_indices = None  # the original indices matched to the new, with entry -1 when flat
        self._original_raw_indices = None
        self._formatted_subset_definition = None
        self._raw_subset_definition = None
        self._squeeze = squeeze
        self._parent = parent

        raw_shape, formatted_shape = self._validate_subset_definition(
            subset_definition, coordinate_basis)
        DataSegment.__init__(
            self, parent.raw_dtype, raw_shape, parent.formatted_dtype, formatted_shape,
            mode=parent.mode)
        self._pixels_written = 0
        if self.mode == 'w':
            self._expected_pixels_written = int(numpy.prod(raw_shape))
        else:
            self._expected_pixels_written = 0

    def _validate_shapes(self) -> None:
        # handled else where
        pass

    @property
    def parent(self) -> DataSegment:
        return self._parent

    @property
    def formatted_subset_definition(self) -> Tuple[slice, ...]:
        """
        Tuple[slice]: The subset definition, in formatted coordinates.
        """

        return self._formatted_subset_definition

    @property
    def raw_subset_definition(self) -> Tuple[slice, ...]:
        """
        Tuple[slice]: The subset definition, in raw coordinates.
        """

        return self._raw_subset_definition

    @property
    def close_parent(self) -> bool:
        """
        bool: Call parent.close() when close is called?
        """

        return self._close_parent

    @close_parent.setter
    def close_parent(self, value):
        self._close_parent = bool(value)

    def _validate_subset_definition(
            self,
            subset_definition: Tuple[slice, ...],
            coordinate_basis: str) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """
        Validates the subset definition.

        Parameters
        ----------
        subset_definition : Tuple[slice, ...]
        coordinate_basis : str

        Returns
        -------
        raw_shape : Tuple[int, ...]
        formatted_shape : Tuple[int, ...]
        """

        raw_shape = []
        formatted_shape = []
        original_indices = []
        raw_indices = []

        coordinate_basis = coordinate_basis.strip().lower()
        if coordinate_basis == 'raw':
            raw_def = self.parent.verify_raw_subscript(subset_definition)
            form_def = self.parent.format_function.transform_raw_slice(raw_def)
        elif coordinate_basis == 'formatted':
            form_def = self.parent.verify_formatted_subscript(subset_definition)
            raw_def = self.parent.format_function.transform_formatted_slice(form_def)
        else:
            raise ValueError('Got unexpected coordinate basis `{}`'.format(coordinate_basis))

        for index, entry in enumerate(form_def):
            this_size = get_slice_result_size(entry)
            if self._squeeze and this_size == 1:
                logger.info('Entry at index {} of subset definition yields a single entry'.format(index))
                original_indices.append(-1)
            else:
                formatted_shape.append(this_size)
                original_indices.append(index)

        self._formatted_subset_definition = form_def
        self._raw_subset_definition = raw_def
        self._original_formatted_indices = tuple(original_indices)

        for index, entry in enumerate(raw_def):
            this_size = get_slice_result_size(entry)
            if self._squeeze and this_size == 1:
                logger.info('Raw slice at index {} of subset definition yields a single entry'.format(index))
                raw_indices.append(-1)
            else:
                raw_shape.append(this_size)
                raw_indices.append(index)
        self._original_raw_indices = tuple(raw_indices)
        return tuple(raw_shape), tuple(formatted_shape)

    def _get_parent_subscript(
            self,
            norm_subscript: Tuple[slice, ...],
            this_shape: Tuple[int, ...],
            full_shape: Tuple[int, ...],
            use_indices: Tuple[int, ...],
            subset_definition: Tuple[slice, ...]) -> Tuple[slice, ...]:
        """
        Helper function for defining a parent subscript from the subset subscript definition.

        Parameters
        ----------
        norm_subscript : Tuple[slice, ...]
            The normalized subset subscript.
        this_shape : Tuple[int, ...]
            The shape in the subset domain.
        full_shape : Tuple[int, ...]
            The full parent shape in the given domain.
        use_indices : Tuple[int, ...]
            Structure helping to identify the dimensions from the parent which
            have been preserved, and which have collapsed.
        subset_definition : Tuple[slice, ...]
            The subset definition with respect to parent coordinates.

        Returns
        -------
        Tuple[slice, ...]
        """

        out = []
        for full_size, out_index, slice_def in zip(full_shape, use_indices, subset_definition):
            if out_index == -1:
                out.append(slice_def)
            else:
                part_def = norm_subscript[out_index]
                step = part_def.step*slice_def.step
                # now, extract start and stop
                # the logic of this is terrible, so let's just do it the easy way
                the_array = numpy.arange(full_size)[slice_def][part_def]
                if len(the_array) < 1:
                    raise KeyError('Got invalid slice definition {} for shape {}'.format(norm_subscript, this_shape))
                start = the_array[0]
                stop = the_array[-1] + step
                if stop < 0:
                    stop = None
                elif stop > full_size:
                    stop = full_size
                out.append(slice(start, stop, step))
        return tuple(out)

    def get_parent_raw_subscript(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice]]]) -> Tuple[slice, ...]:
        """
        Gets the raw parent subscript from the raw subset subscript definition.

        Parameters
        ----------
        subscript : None|int|slice|Sequence[int|slice]

        Returns
        -------
        Tuple[slice, ...]
        """

        return self._get_parent_subscript(
            self.verify_raw_subscript(subscript), self.raw_shape, self.parent.raw_shape,
            self._original_raw_indices, self._raw_subset_definition)

    def get_parent_formatted_subscript(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice]]]) -> Tuple[slice, ...]:
        """
        Gets the formatted parent subscript from the formatted subset subscript definition.

        Parameters
        ----------
        subscript : None|int|slice|Sequence[int|slice]

        Returns
        -------
        Tuple[slice, ...]
        """

        out = self._get_parent_subscript(
            self.verify_formatted_subscript(subscript), self.formatted_shape, self.parent.formatted_shape,
            self._original_formatted_indices, self._formatted_subset_definition)
        return out

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        norm_subscript = self.get_parent_raw_subscript(subscript)
        if squeeze:
            return self.parent.read_raw(norm_subscript, squeeze=True)
        else:
            data = self.parent.read_raw(norm_subscript, squeeze=False)
            use_shape = []
            for check, size in zip(self._original_raw_indices, data.shape):
                if not self._squeeze or check != -1:
                    use_shape.append(size)
            if squeeze:
                return numpy.squeeze(data)
            else:
                return numpy.reshape(data, tuple(use_shape))

    def read(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        norm_subscript = self.get_parent_formatted_subscript(subscript)
        if squeeze:
            return self.parent.read(norm_subscript, squeeze=True)
        else:
            data = self.parent.read(norm_subscript, squeeze=False)
            use_shape = []
            for check, size in zip(self._original_formatted_indices, data.shape):
                if not self._squeeze or check != -1:
                    use_shape.append(size)
            if squeeze:
                return numpy.squeeze(data)
            else:
                return numpy.reshape(data, tuple(use_shape))

    def check_fully_written(self, warn: bool = False) -> bool:
        if self.mode == 'r':
            return True

        if self._pixels_written < self._expected_pixels_written:
            if warn:
                logger.error(
                    'Segment expected {} pixels written, but only {} pixels were written'.format(
                        self._expected_pixels_written, self._pixels_written))
            return False
        elif self._pixels_written == self._expected_pixels_written:
            return True
        else:
            if warn:
                logger.error(
                    'Segment expected {} pixels written,\n\t'
                    'but {} pixels were written.\n\t'
                    'This redundancy may be an error'.format(
                        self._expected_pixels_written, self._pixels_written))
            return False

    def _update_pixels_written(self, written: int) -> None:
        new_pixels_written = self._pixels_written + written
        if self._pixels_written <= self._expected_pixels_written < new_pixels_written:
            logger.error(
                'Segment expected {} pixels written,\n\t'
                'but now has {} pixels written.\n\t'
                'This redundancy may be an error'.format(
                    self._expected_pixels_written, new_pixels_written))
        self._pixels_written = new_pixels_written

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs):
        self._validate_closed()
        self._verify_write_raw_details(data)
        subscript = _infer_subscript_for_write(data, start_indices, subscript, self.raw_shape)
        parent_subscript = self.get_parent_raw_subscript(subscript)
        self.parent.write_raw(data, subscript=parent_subscript, **kwargs)
        self._update_pixels_written(data.size)

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        """
        This returns the bytes for the underlying raw data **of the parent segment.**

        The only writing use case considered at present is for the blocks including
        padding inside a NITF file.

        Parameters
        ----------
        warn : bool
            If `True`, then a check will be performed to ensure that the data
            has been fully written.

        Returns
        -------
        bytes|Tuple
            The result will be a `bytes` object, unless the data segment is
            made up of a collection of child data segments, in which case the
            result will be a Tuple consisting of their `get_raw_bytes` returns.
        """

        self._validate_closed()
        if warn and not self.check_fully_written(warn=True):
            logger.error(
                'There has been a call to `get_raw_bytes` from {},\n\t'
                'but all pixels are not fully written'.format(self.__class__))
        return self.parent.get_raw_bytes(warn=False)

    def flush(self) -> None:
        self._validate_closed()
        try:
            self.parent.flush()
        except AttributeError:
            return

    def close(self):
        try:
            if self._closed:
                return

            self.flush()
            if self.close_parent:
                self.parent.close()
            DataSegment.close(self)
            self._parent = None
        except (ValueError, AttributeError):
            return


class BandAggregateSegment(DataSegment):
    """
    This stacks a collection of data segments, which must have compatible details,
    together along a new (final) band dimension.

    Note that :func:`read` and :func:`read_raw` return identical results here.
    To access raw data from the children, use access on the `children` property.

    Introduced in version 1.3.0.
    """

    __slots__ = ('_band_dimension', '_children', '_close_children')

    def __init__(
            self,
            children: Sequence[DataSegment],
            band_dimension: int,
            formatted_dtype: Optional[Union[str, numpy.dtype]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            close_children: bool = True):
        """

        Parameters
        ----------
        children : Sequence[DataSegment]
        band_dimension : int
            The band dimension. This must remain unchanged in transpose_axes,
            and is not permitted to be reversed by reverse_axes.
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform, after applying any axis reversal,
            and before applying any format function
        format_function : None|FormatFunction
        close_children : bool
        """

        self._band_dimension = None
        self._close_children = None
        self.close_children = close_children
        self._children = None
        self._set_band_dimension(band_dimension, reverse_axes, transpose_axes)
        raw_dtype, raw_shape, form_shape, the_mode = self._set_children(children, transpose_axes)

        if format_function is None:
            formatted_dtype = raw_dtype
            formatted_shape = form_shape
        else:
            if formatted_dtype is None or formatted_shape is None:
                raise ValueError(
                    'If format_function is provided,\n\t'
                    'then formatted_dtype and formatted_shape must be provided.')
        DataSegment.__init__(
            self, raw_dtype, raw_shape, formatted_dtype, formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode=the_mode)

    @property
    def band_dimension(self) -> int:
        """
        int: The band dimension, in raw data after the transpose operation.
        """

        return self._band_dimension

    def _set_band_dimension(
            self,
            value: int,
            reverse_axes: Union[None, int, Sequence[int]],
            transpose_axes: Union[None, Tuple[int, ...]]) -> None:
        if not isinstance(value, int):
            raise TypeError('band_dimension must be an integer')
        if value < 0:
            raise TypeError('band_dimension must be non-negative')

        if transpose_axes is not None:
            if value != transpose_axes[value]:
                raise ValueError('band_dimension is not permitted to be changed by transpose_axes.')

        if reverse_axes is None:
            pass
        elif isinstance(reverse_axes, int):
            if value == reverse_axes:
                raise ValueError('Reversal along the band dimension is not permitted')
        else:
            if value in reverse_axes:
                raise ValueError('Reversal along the band dimension is not permitted')

        if self._band_dimension is not None:
            if value != self._band_dimension:
                raise ValueError('band_dimension is read only once set')
            return  # nothing to be done
        self._band_dimension = value

    @property
    def close_children(self) -> bool:
        """
        bool: Call child.close() when close is called?
        """

        return self._close_children

    @close_children.setter
    def close_children(self, value):
        self._close_children = bool(value)

    @property
    def children(self) -> Tuple[DataSegment, ...]:
        """
        The collection of children that we are stacking.

        Returns
        -------
        Tuple[DataSegment, ...]
        """

        return self._children

    def _set_children(
            self,
            children: Sequence[DataSegment],
            transpose_axes: Optional[Tuple[int, ...]]) -> Tuple[numpy.dtype, Tuple[int, ...], Tuple[int, ...], str]:
        if len(children) < 2:
            raise ValueError('Cannot define a BandAggregateSegment based on fewer than 2 segments.')

        child_shape = children[0].formatted_shape
        the_dtype = children[0].formatted_dtype
        the_mode = children[0].mode

        if transpose_axes is None:
            transpose_axes = tuple(range(0, len(child_shape) + 1))

        raw_shape = [entry for entry in child_shape]
        raw_shape.insert(self.band_dimension, len(children))
        raw_shape = tuple(raw_shape)

        form_shape = tuple(raw_shape[entry] for entry in transpose_axes)

        use_children = []
        for child in children:
            if child.formatted_shape != child_shape:
                raise ValueError('All children must have the same formatted shape')
            if child.formatted_dtype != the_dtype:
                raise ValueError('All children must have the same formatted dtype')
            if child.mode != the_mode:
                raise ValueError('All children must have the same mode')
            if child.mode == 'w' and not child.can_write_regular:
                raise ValueError('write mode requires that all children can write regular data')
            use_children.append(child)
        self._children = tuple(use_children)
        return the_dtype, raw_shape, form_shape, the_mode

    @property
    def bands(self) -> int:
        """
        int: The number of bands (child data segments)
        """

        return len(self.children)

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        norm_subscript, the_shape = get_subscript_result_size(subscript, self.raw_shape)
        out = numpy.empty(the_shape, dtype=self.raw_dtype)
        full_band_subscript = tuple(slice(0, entry, 1) for entry in the_shape)

        for out_index, index in enumerate(numpy.arange(self.bands)[norm_subscript[self.band_dimension]]):
            child_subscript = norm_subscript[:self.band_dimension] + norm_subscript[self.band_dimension+1:]
            band_subscript = full_band_subscript[:self.band_dimension] + \
                (out_index, ) + \
                full_band_subscript[self.band_dimension+1:]
            out[band_subscript] = self.children[index].read(child_subscript, squeeze=False)

        if squeeze:
            return numpy.squeeze(out)
        else:
            return out

    def check_fully_written(self, warn: bool = False) -> bool:
        if self.mode == 'r':
            return True

        out = True
        for i, child in enumerate(self.children):
            done = child.check_fully_written(warn=warn)
            if warn and not done:
                logger.error('Band {} of BandAggregateSegment indicates incomplete writing'.format(i))
            out &= done
        return out

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs):
        """
        In keeping with data segment mode, write the data provided in raw form,
        assuming the slice specified relative to raw data coordinates. This
        requires that `mode` is `'w'`.

        Note that raw form/order for the data segment is simply a reordered version
        of the formatted data for parent. This **is not** related to raw data
        with respect to the parent. To write raw data with respect to the parent,
        use :func:`parent.write_raw` instead.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data in raw form.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Sequence[slice]
            The subscript definition in raw coordinates.
        kwargs

        Returns
        -------
        None
        """

        self._validate_closed()
        self._verify_write_raw_details(data)

        norm_subscript = _infer_subscript_for_write(
            data, start_indices, subscript, self.raw_shape)
        # iterate over each band, and write the raw data there...
        for out_index, index in enumerate(
                numpy.arange(self.bands)[norm_subscript[self.band_dimension]]):
            child_subscript = norm_subscript[:self.band_dimension] + \
                norm_subscript[self.band_dimension+1:]
            band_subscript = norm_subscript[:self.band_dimension] + \
                (out_index, ) + \
                norm_subscript[self.band_dimension+1:]
            self.children[index].write(
                data[band_subscript], subscript=child_subscript, **kwargs)

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        self._validate_closed()
        return tuple(entry.get_raw_bytes(warn=warn) for entry in self.children)

    def flush(self) -> None:
        self._validate_closed()
        try:
            if self.children is not None:
                for child in self.children:
                    child.flush()
        except AttributeError:
            return

    def close(self):
        try:
            if self._closed:
                return

            self.flush()
            if self._children is not None:
                for entry in self._children:
                    entry.close()
            DataSegment.close(self)
            self._children = None
        except AttributeError:
            return


class BlockAggregateSegment(DataSegment):
    """
    Combines a collection of child data segments, according to a given
    block definition. All children must have the same formatted_dtype. This
    implementation is motivated by a two-dimensional block arrangement, but is
    entirely general.

    No effort is made to ensure that the block definition spans the whole space,
    nor is any effort made to ensure that blocks do not overlap.

    If there are holes present in block definition, then data read across any
    hole will be populated with `missing_data_value`. Data attempted to write
    across any hole will simply be ignored.

    Introduced in version 1.3.0.
    """

    __slots__ = (
        '_children', '_formatted_child_arrangement', '_raw_child_arrangement',
        '_missing_data_value', '_close_children')

    def __init__(
            self,
            children: Sequence[DataSegment],
            child_arrangement: Sequence[Tuple[slice, ...]],
            coordinate_basis: str,
            missing_data_value,
            raw_shape: Tuple[int, ...],
            formatted_dtype: Union[str, numpy.dtype],
            formatted_shape: Tuple[int, ...],
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            close_children: bool = True):
        """

        Parameters
        ----------
        children : Sequence[DataSegment]
            The collection of children
        child_arrangement : Sequence[Tuple[slice, ...]]
            The collection of definitions for each block. Overlap in definition
            is permitted.
        missing_data_value
            Missing data value, which must be compatible with
            raw_dtype=child.formatted_dtype.
        close_children : bool
        """

        self.close_children = close_children
        self._close_children = None

        self._children = None
        self._formatted_child_arrangement = None
        self._raw_child_arrangement = None
        self._missing_data_value = missing_data_value
        raw_dtype = children[0].formatted_dtype
        the_mode = children[0].mode
        DataSegment.__init__(
            self, raw_dtype, raw_shape, formatted_dtype, formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode=the_mode)
        self._set_children(children, child_arrangement, coordinate_basis)

    @property
    def close_children(self) -> bool:
        """
        bool: Call child.close() when close is called?
        """

        return self._close_children

    @close_children.setter
    def close_children(self, value):
        self._close_children = bool(value)

    @property
    def children(self) -> Tuple[DataSegment, ...]:
        """
        The collection of children that we are stacking together.

        Returns
        -------
        Tuple[DataSegment, ...]
        """

        return self._children

    def _set_children(
            self,
            children: Sequence[DataSegment],
            child_arrangement: Sequence[Tuple[slice, ...]],
            coordinate_basis: str) -> None:

        if len(children) != len(child_arrangement):
            raise ValueError('We must have the same number of children as child_arrangement entries')

        coordinate_basis = coordinate_basis.strip().lower()
        if coordinate_basis == 'raw':
            raw_arrangement = [self.verify_raw_subscript(entry) for entry in child_arrangement]
            formatted_arrangement = [self.format_function.transform_raw_slice(entry) for entry in raw_arrangement]
        elif coordinate_basis == 'formatted':
            formatted_arrangement = [self.verify_formatted_subscript(entry) for entry in child_arrangement]
            raw_arrangement = [self.format_function.transform_formatted_slice(entry) for entry in formatted_arrangement]
        else:
            raise ValueError('Got unexpected coordinate basis `{}`'.format(coordinate_basis))

        for i, (child, raw_def, form_def) in enumerate(zip(children, raw_arrangement, formatted_arrangement)):
            if child.formatted_dtype != self.raw_dtype:
                raise ValueError(
                    'Each child.formatted_dtype must be identical to\n\t'
                    'self.raw_dtype = {}'.format(self.raw_dtype))

            for entry in raw_def:
                if entry.step not in [1, -1]:
                    raise ValueError('Each entry of child_arrangement must have step +/-1.')
            for entry in form_def:
                if entry.step not in [1, -1]:
                    raise ValueError('Each entry of child_arrangement must have step +/-1.')

            # verify the shape is sensible
            _, result_shape = get_subscript_result_size(raw_def, self.raw_shape)
            if result_shape != child.formatted_shape:
                raise ValueError(
                    'child_arrangement definition expects child {} to have formatted_shape {},\n\t'
                    'but it has formatted_shape {}'.format(i, result_shape, child.formatted_shape))

        self._children = tuple(children)
        self._raw_child_arrangement = tuple(raw_arrangement)
        self._formatted_child_arrangement = tuple(formatted_arrangement)

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        subscript, formatted_shape = get_subscript_result_size(subscript, self.raw_shape)
        out = numpy.full(formatted_shape, fill_value=self._missing_data_value, dtype=self.raw_dtype)

        for entry, child in zip(self._raw_child_arrangement, self._children):
            use_block = True
            parent_subscript = []
            child_subscript = []
            for data_slice, block_slice in zip(subscript, entry):
                child_entry, par_entry = _find_slice_overlap(data_slice, block_slice)
                if par_entry is None:
                    use_block = False
                    # there is no overlap
                    break
                else:
                    parent_subscript.append(par_entry)
                    child_subscript.append(child_entry)
            if use_block:
                out[tuple(parent_subscript)] = child.read_raw(tuple(child_subscript), squeeze=False)

        if squeeze:
            return numpy.squeeze(out)
        else:
            return out

    def check_fully_written(self, warn: bool = False) -> bool:
        if self.mode == 'r':
            return True

        out = True
        for i, child in enumerate(self.children):
            done = child.check_fully_written(warn=warn)
            if warn and not done:
                logger.error('Block {} of BlockAggregateSegment indicates incomplete writing'.format(i))
            out &= done
        return out

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs):
        """
        In keeping with data segment mode, write the data provided in raw form,
        assuming the slice specified relative to raw data coordinates. This
        requires that `mode` is `'w'`.

        Note that raw form/order for the data segment is broken into blocks of
        reordered version of the formatted data for each child. This **is not**
        related to raw data with respect to the child. To write raw data with
        respect to the parent, use :func:`child.write_raw` instead.

        **Only one of `start_indices` and `subscript` should be specified.**

        Parameters
        ----------
        data : numpy.ndarray
            The data in raw form.
        start_indices : None|int|Tuple[int, ...]
            Assuming a contiguous chunk of data, this provides the starting
            indices of the chunk. Any missing (tail) coordinates will be filled
            in with 0's.
        subscript : None|Sequence[slice]
            The subscript definition in raw coordinates.
        kwargs

        Returns
        -------
        None
        """

        self._validate_closed()
        self._verify_write_raw_details(data)

        norm_subscript = _infer_subscript_for_write(data, start_indices, subscript, self.raw_shape)
        for entry, child in zip(self._raw_child_arrangement, self._children):
            # determine if there is overlap of norm_subscript with this block,
            # and write the appropriate data, if so.
            use_block = True
            data_subscript = []
            child_subscript = []
            for lim, data_slice, block_slice in zip(self.raw_shape, norm_subscript, entry):
                child_entry, par_entry = _find_slice_overlap(data_slice, block_slice)
                # this expresses the overlap between our data slice in overall coordinates
                #   relative to the entire raw_shape, and coordinates relative to
                #   just the block in question
                if par_entry is None:
                    use_block = False
                    # there is no overlap
                    break

                # we need to convert from overall coordinates relative to the entire
                # raw image, to coordinates relative to just the data shape
                _, data_entry = _find_slice_overlap(slice(0, lim, 1), par_entry)
                data_subscript.append(data_entry)
                child_subscript.append(child_entry)

            if use_block:
                child.write(data[tuple(data_subscript)], subscript=tuple(child_subscript), **kwargs)

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        self._validate_closed()
        return tuple(entry.get_raw_bytes(warn=warn) for entry in self.children)

    def flush(self) -> None:
        self._validate_closed()
        try:
            if self.children is not None:
                for child in self.children:
                    child.flush()
        except AttributeError:
            return

    def close(self):
        try:
            if self._closed:
                return

            self.flush()
            if self._children is not None:
                for entry in self._children:
                    entry.close()
            DataSegment.close(self)
            self._children = None
        except AttributeError:
            return


####
# Concrete implementations

class NumpyArraySegment(DataSegment):
    """
    DataSegment based on reading from a numpy.ndarray.

    Introduced in version 1.3.0.
    """

    __slots__ = ('_underlying_array', '_pixels_written', '_expected_pixels_written')

    def __init__(
            self,
            underlying_array: numpy.ndarray,
            formatted_dtype: Optional[Union[str, numpy.dtype]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            mode: str = 'r'):
        """

        Parameters
        ----------
        underlying_array : numpy.ndarray
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform to the raw data, after applying
            any axis reversal, and before applying any format function
        format_function : None|FormatFunction
        mode : str
        """

        if not isinstance(underlying_array, numpy.ndarray):
            raise TypeError(
                'underlying array must be a numpy.ndarray, got type `{}`'.format(
                    type(underlying_array)))
        self._underlying_array = underlying_array
        self._pixels_written = 0
        if formatted_dtype is None:
            if format_function is None:
                formatted_dtype = underlying_array.dtype
            else:
                raise ValueError(
                    'Format function is provided, so formatted_dtype must be provided.')
        if formatted_shape is None:
            if format_function is None:
                if transpose_axes is None:
                    formatted_shape = underlying_array.shape
                else:
                    formatted_shape = [underlying_array.shape[index] for index in transpose_axes]
            else:
                raise ValueError(
                    'Format function is provided, so formatted_shape must be provided.')

        DataSegment.__init__(
            self, underlying_array.dtype, underlying_array.shape, formatted_dtype, formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes, format_function=format_function,
            mode=mode)
        if self.mode == 'w':
            self._expected_pixels_written = self._underlying_array.size
        else:
            self._expected_pixels_written = 0

    @property
    def underlying_array(self) -> numpy.ndarray:
        """
        The underlying data array.

        Returns
        -------
        numpy.ndarray
        """

        return self._underlying_array

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        if self.mode != 'r':
            raise ValueError('Requires mode == "r"')

        subscript, out_shape = get_subscript_result_size(subscript, self.raw_shape)
        out = self._underlying_array[subscript]  # squeezed by default

        if squeeze:
            return numpy.squeeze(out)
        else:
            return numpy.reshape(out, out_shape)

    def check_fully_written(self, warn: bool = False) -> bool:
        if self.mode == 'r':
            return True

        if self._pixels_written < self._expected_pixels_written:
            if warn:
                logger.error(
                    'Segment expected {} pixels written, but only {} pixels were written'.format(
                        self._expected_pixels_written, self._pixels_written))
            return False
        elif self._pixels_written == self._expected_pixels_written:
            return True
        else:
            if warn:
                logger.error(
                    'Segment expected {} pixels written,\n\t'
                    'but {} pixels were written.\n\t'
                    'This redundancy may be an error'.format(
                        self._expected_pixels_written, self._pixels_written))
            return False

    def _update_pixels_written(self, written: int) -> None:
        new_pixels_written = self._pixels_written + written
        if self._pixels_written <= self._expected_pixels_written < new_pixels_written:
            logger.error(
                'Segment expected {} pixels written,\n\t'
                'but now has {} pixels written.\n\t'
                'This redundancy may be an error'.format(
                    self._expected_pixels_written, new_pixels_written))
        self._pixels_written = new_pixels_written

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Optional[Union[int, Tuple[int, ...]]] = None,
            subscript: Optional[Sequence[slice]] = None,
            **kwargs):
        self._validate_closed()
        self._verify_write_raw_details(data)
        subscript = _infer_subscript_for_write(data, start_indices, subscript, self.raw_shape)
        self._underlying_array[subscript] = data
        self._update_pixels_written(data.size)

    def get_raw_bytes(self, warn: bool = False) -> Union[bytes, Tuple]:
        self._validate_closed()
        if warn and not self.check_fully_written(warn=True):
            logger.error(
                'There has been a call to `get_raw_bytes` from {},\n\t'
                'but all pixels are not fully written'.format(self.__class__))
        return self.underlying_array.tobytes()

    def flush(self) -> None:
        self._validate_closed()
        try:
            if self.mode == 'w' and hasattr(self._underlying_array, 'flush'):
                # noinspection PyUnresolvedReferences
                self._underlying_array.flush()
        except AttributeError:
            return

    def close(self) -> None:
        try:
            if self._closed:
                return

            self.flush()
            self._underlying_array = None
            DataSegment.close(self)
        except AttributeError:
            return


class NumpyMemmapSegment(NumpyArraySegment):
    """
    DataSegment based on establishing a numpy memmap, and using that as the
    underlying array.

    Introduced in version 1.3.0.
    """

    __slots__ = (
        '_file_object', '_memory_map', '_close_file')

    def __init__(
            self,
            file_object: Union[str, BinaryIO],
            data_offset: int,
            raw_dtype: Union[str, numpy.dtype],
            raw_shape: Tuple[int, ...],
            formatted_dtype: Optional[Union[str, numpy.dtype]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            mode: str = 'r',
            close_file: bool = False):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
        data_offset : int
        raw_dtype : str|numpy.dtype
        raw_shape : Tuple[int, ...]
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform to the raw data, after applying
            any axis reversal, and before applying any format function
        format_function : None|FormatFunction
        mode : str
        close_file : bool
        """

        self._close_file = None
        if isinstance(file_object, str):
            close_file = True
        self.close_file = close_file
        self._file_object = file_object

        self._pixels_written = 0
        self._expected_pixels_written = 0

        mmap_mode = 'r' if mode == 'r' else 'r+'
        self._memory_map = numpy.memmap(
            file_object,
            offset=data_offset,
            dtype=raw_dtype,
            shape=raw_shape,
            mode=mmap_mode)

        NumpyArraySegment.__init__(
            self, self._memory_map, formatted_dtype=formatted_dtype, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode=mode)

    @property
    def close_file(self) -> bool:
        """
        bool: Close the file object when complete?
        """

        return self._close_file

    @close_file.setter
    def close_file(self, value):
        self._close_file = bool(value)

    def flush(self) -> None:
        try:
            if self.mode == 'w':
                self._memory_map.flush()
        except AttributeError:
            pass

    def close(self):
        try:
            if self._closed:
                return

            NumpyArraySegment.close(self)  # NB: this calls flush
            self._memory_map = None
            if self._close_file and hasattr(self._file_object, 'close'):
                self._file_object.close()
            self._file_object = None
        except AttributeError:
            return


class HDF5DatasetSegment(DataSegment):
    """
    DataSegment based on reading from an hdf5 file, using the h5py library.

    Introduced in version 1.3.0.
    """
    _allowed_modes = ('r', )

    __slots__ = (
        '_file_object', '_data_set', '_close_file')

    def __init__(
            self,
            file_object: Union[str, h5pyFile],
            data_set: Union[str, h5pyDataset],
            formatted_dtype: Optional[Union[str, numpy.dtype]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            close_file: bool = False):
        """

        Parameters
        ----------
        file_object : str|h5py.File
        data_set : str|h5py.Dataset
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform to the raw data, after applying
            any axis reversal, and before applying any format function
        format_function : None|FormatFunction
        close_file : bool
        """

        self._close_file = None
        self._file_object = None
        self._data_set = None

        if h5py is None:
            raise ValueError(
                'h5py was not successfully imported, and no hdf5 file can be read')

        if isinstance(file_object, str):
            close_file = True

        self._set_file_object(file_object)
        self._set_data_set(data_set)

        if format_function is not None:
            if formatted_dtype is None or formatted_shape is None:
                raise ValueError(
                    'format_function is supplied, so formatted_dtype and formatted_shape must also be supplied')
        else:
            formatted_dtype = self.data_set.dtype
            raw_shape = self.data_set.shape
            if transpose_axes is None:
                formatted_shape = raw_shape
            else:
                formatted_shape = tuple(raw_shape[entry] for entry in transpose_axes)
        self.close_file = close_file

        DataSegment.__init__(
            self, self.data_set.dtype, self.data_set.shape, formatted_dtype, formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode='r')

    @property
    def close_file(self) -> bool:
        """
        bool: Close the file object when complete?
        """

        return self._close_file

    @close_file.setter
    def close_file(self, value):
        self._close_file = bool(value)

    @property
    def file_object(self) -> h5pyFile:
        return self._file_object

    def _set_file_object(self, value) -> None:
        if isinstance(value, str):
            value = h5py.File(value, mode='r')
        if not isinstance(value, h5py.File):
            raise ValueError('Requires a path to a hdf5 file or h5py.File object')
        self._file_object = value

    @property
    def data_set(self) -> h5pyDataset:
        return self._data_set

    def _set_data_set(self, value) -> None:
        if isinstance(value, str):
            value = self.file_object[value]
        if not isinstance(value, h5py.Dataset):
            raise ValueError('Requires a dataset path or h5py.Dataset object')
        self._data_set = value

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        subscript, out_shape = get_subscript_result_size(subscript, self.raw_shape)

        # NB: h5py does not support slicing with a negative step (right now)
        #   we need to read the identical elements in positive order,
        #   then reverse. Note that this is not the same as using the mirror
        #   image slice, which is used in the reverse_axes operations.

        reverse = []
        use_subscript = []
        for index, (the_size, entry) in enumerate(zip(self.raw_shape, subscript)):
            if entry.step < 0:
                use_subscript.append(_reverse_slice(entry))
                reverse.append(index)
            else:
                use_subscript.append(entry)
        use_subscript = tuple(use_subscript)

        out = numpy.reshape(self.data_set[use_subscript], out_shape)
        for index in reverse:
            out = numpy.flip(out, axis=index)

        if squeeze:
            return numpy.squeeze(out)
        else:
            return out

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs):
        raise NotImplementedError

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        raise NotImplementedError

    def check_fully_written(self, warn: bool = False) -> bool:
        return True

    def close(self) -> None:
        try:
            if self._closed:
                return

            self._data_set = None
            if self._close_file and hasattr(self.file_object, 'close'):
                self.file_object.close()
            self._file_object = None
            DataSegment.close(self)
        except AttributeError:
            pass


class FileReadDataSegment(DataSegment):
    """
    Read a data array manually from a file - this is primarily intended for cloud
    usage.

    Introduced in version 1.3.0.
    """
    _allowed_modes = ('r', )

    __slots__ = (
        '_file_object', '_data_offset', '_close_file')

    def __init__(
            self,
            file_object: BinaryIO,
            data_offset: int,
            raw_dtype: Union[str, numpy.dtype],
            raw_shape: Tuple[int, ...],
            formatted_dtype: Union[str, numpy.dtype],
            formatted_shape: Tuple[int, ...],
            reverse_axes: Optional[Union[int, Sequence[int]]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            format_function: Optional[FormatFunction] = None,
            close_file: bool = False):
        """

        Parameters
        ----------
        file_object : BinaryIO
        data_offset : int
        raw_dtype : str|numpy.dtype
        raw_shape : Tuple[int, ...]
        formatted_dtype : str|numpy.dtype
        formatted_shape : Tuple[int, ...]
        reverse_axes : None|int|Sequence[int]
            The collection of axes (in raw order) to reverse, prior to applying
            transpose operation
        transpose_axes : None|Tuple[int, ...]
            The transpose operation to perform to the raw data, after applying
            any axis reversal, and before applying any format function
        format_function : None|FormatFunction
        close_file : bool
        """

        self._file_object = None
        self._data_offset = None
        self._close_file = None
        self.close_file = close_file
        self._set_data_offset(data_offset)
        self._set_file_object(file_object)
        DataSegment.__init__(
            self, raw_dtype, raw_shape, formatted_dtype, formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode='r')

    @property
    def close_file(self) -> bool:
        """
        bool: Close the file object when complete?
        """

        return self._close_file

    @close_file.setter
    def close_file(self, value):
        self._close_file = bool(value)

    @property
    def file_object(self) -> BinaryIO:
        return self._file_object

    def _set_file_object(self, value) -> None:
        if not is_file_like(value):
            raise ValueError('Requires a file-like object')
        self._file_object = value

    @property
    def data_offset(self) -> int:
        """
        int: The offset of the data in bytes from the start of the file-like
        object.
        """

        return self._data_offset

    def _set_data_offset(self, value: int) -> None:
        value = int(value)
        if value < 0:
            raise ValueError('data_offset must be non-negative.')
        self._data_offset = value

    def read_raw(
            self,
            subscript: Union[None, int, slice, Sequence[Union[int, slice, Tuple[int, ...]]]],
            squeeze=True) -> numpy.ndarray:
        self._validate_closed()
        subscript, out_shape = get_subscript_result_size(subscript, self.raw_shape)

        init_slice = subscript[0]
        init_reverse = (init_slice.step < 0)
        if init_reverse:
            init_slice = _reverse_slice(init_slice)

        pixel_per_row = 1 if self.formatted_ndim == 1 else int(numpy.prod(self.raw_shape[1:]))
        row_stride = self.raw_dtype.itemsize*pixel_per_row

        start_row = init_slice.start
        rows = init_slice.stop - init_slice.start

        # read the whole contiguous chunk from start_row up to the final row
        # seek to the proper start location
        start_loc = self._data_offset + start_row*row_stride
        self.file_object.seek(start_loc, os.SEEK_SET)
        total_size = rows*row_stride
        # read our data
        data = self.file_object.read(total_size)
        if len(data) != total_size:
            raise ValueError(
                'Tried to read {} bytes of data, but received {}.\n'
                'The most likely reason for this is a malformed chipper, \n'
                'which attempts to read more data than the file contains'.format(total_size, len(data)))
        # define temp array from this data
        data = numpy.frombuffer(data, self._raw_dtype, rows*pixel_per_row)
        data = numpy.reshape(data, (rows, ) + self.raw_shape[1:])
        # extract our data
        out = data[(slice(None, None, init_slice.step), ) + subscript[1:]]
        out = numpy.reshape(out, out_shape)
        if init_reverse:
            out = numpy.flip(out, axis=0)

        if squeeze:
            out = numpy.copy(numpy.squeeze(out))
        else:
            out = numpy.copy(out)
        del data
        return out

    def write_raw(
            self,
            data: numpy.ndarray,
            start_indices: Union[None, int, Tuple[int, ...]] = None,
            subscript: Union[None, Sequence[slice]] = None,
            **kwargs):

        if self.mode != 'w':
            raise ValueError('I/O Error, functionality requires mode == "w"')
        raise NotImplementedError

    def get_raw_bytes(self, warn: bool = True) -> Union[bytes, Tuple]:
        raise NotImplementedError

    def check_fully_written(self, warn: bool = False) -> bool:
        return True

    def close(self) -> None:
        try:
            if self._closed:
                return

            if self._close_file:
                if hasattr(self.file_object, 'close'):
                    self.file_object.close()
            self._file_object = None
            DataSegment.close(self)
        except AttributeError:
            return



"""
The main NITF header definitions.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging

from .base import NITFElement, UserHeaderType, _IntegerDescriptor,\
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor, _RawDescriptor, \
    _ItemArrayHeaders
from .security import NITFSecurityTags, NITFSecurityTags0

logger = logging.getLogger(__name__)


#############
# NITF 2.1 version

class ImageSegmentsType(_ItemArrayHeaders):
    """
    This holds the image subheader and item sizes.
    """

    _subhead_len = 6
    _item_len = 10


class GraphicsSegmentsType(_ItemArrayHeaders):
    """
    This holds the graphics subheader and item sizes.
    """

    _subhead_len = 4
    _item_len = 6


class TextSegmentsType(_ItemArrayHeaders):
    """
    This holds the text subheader size and item sizes.
    """

    _subhead_len = 4
    _item_len = 5


class DataExtensionsType(_ItemArrayHeaders):
    """
    This holds the data extension subheader and item sizes.
    """

    _subhead_len = 4
    _item_len = 9


class ReservedExtensionsType(_ItemArrayHeaders):
    """
    This holds the reserved extension subheader and item sizes.
    """

    _subhead_len = 4
    _item_len = 7


class NITFHeader(NITFElement):
    """
    The main NITF file header for NITF version 2.1 - see standards document
    Joint BIIF Profile (JBP) for more information.
    """

    _ordering = (
        'FHDR', 'FVER', 'CLEVEL', 'STYPE',
        'OSTAID', 'FDT', 'FTITLE', 'Security',
        'FSCOP', 'FSCPYS', 'ENCRYP', 'FBKGC',
        'ONAME', 'OPHONE', 'FL', 'HL',
        'ImageSegments', 'GraphicsSegments', 'NUMX',
        'TextSegments', 'DataExtensions', 'ReservedExtensions',
        'UserHeader', 'ExtendedHeader')
    _lengths = {
        'FHDR': 4, 'FVER': 5, 'CLEVEL': 2, 'STYPE': 4,
        'OSTAID': 10, 'FDT': 14, 'FTITLE': 80,
        'FSCOP': 5, 'FSCPYS': 5, 'ENCRYP': 1, 'FBKGC': 3,
        'ONAME': 24, 'OPHONE': 18, 'FL': 12, 'HL': 6,
        'NUMX': 3}
    CLEVEL = _IntegerDescriptor(
        'CLEVEL', True, 2, default_value=0,
        docstring='Complexity Level. This field shall contain the complexity level required to '
                  'interpret fully all components of the file. Valid entries are assigned in '
                  'accordance with complexity levels established in Table A-10.')  # type: int
    STYPE = _StringDescriptor(
        'STYPE', True, 4, default_value='BF01',
        docstring='Standard Type. Standard type or capability. A BCS-A character string `BF01` '
                  'which indicates that this file is formatted using ISO/IEC IS 12087-5. '
                  'NITF02.10 is intended to be registered as a profile of ISO/IEC IS 12087-5.')  # type: str
    OSTAID = _StringDescriptor(
        'OSTAID', True, 10, default_value='',
        docstring='Originating Station ID. This field shall contain the identification code or name of '
                  'the originating organization, system, station, or product. It shall not be '
                  'filled with BCS spaces')  # type: str
    FDT = _StringDescriptor(
        'FDT', True, 14, default_value='',
        docstring='File Date and Time. This field shall contain the time (UTC) of the files '
                  'origination in the format `YYYYMMDDhhmmss`.')  # type: str
    FTITLE = _StringDescriptor(
        'FTITLE', True, 80, default_value='',
        docstring='File Title. This field shall contain the title of the file.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags, default_args={},
        docstring='The image security tags.')  # type: NITFSecurityTags
    FSCOP = _IntegerDescriptor(
        'FSCOP', True, 5, default_value=0,
        docstring='File Copy Number. This field shall contain the copy number of the file.')  # type: int
    FSCPYS = _IntegerDescriptor(
        'FSCPYS', True, 5, default_value=0,
        docstring='File Number of Copies. This field shall contain the total number of '
                  'copies of the file.')  # type: int
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    FBKGC = _RawDescriptor(
        'FBKGC', True, 3, default_value=b'\x00\x00\x00',
        docstring='File Background Color. This field shall contain the three color components of '
                  'the file background in the order Red, Green, Blue.')  # type: bytes
    ONAME = _StringDescriptor(
        'ONAME', True, 24, default_value='',
        docstring='Originator Name. This field shall contain a valid name for the operator '
                  'who originated the file.')  # type: str
    OPHONE = _StringDescriptor(
        'OPHONE', True, 18, default_value='',
        docstring='Originator Phone Number. This field shall contain a valid phone number '
                  'for the operator who originated the file.')  # type: str
    FL = _IntegerDescriptor(
        'FL', True, 12, docstring='The size in bytes of the entire file.')  # type: int
    ImageSegments = _NITFElementDescriptor(
        'ImageSegments', True, ImageSegmentsType, default_args={},
        docstring='The image segment basic information.')  # type: ImageSegmentsType
    GraphicsSegments = _NITFElementDescriptor(
        'GraphicsSegments', True, GraphicsSegmentsType, default_args={},
        docstring='The graphics segment basic information.')  # type: GraphicsSegmentsType
    TextSegments = _NITFElementDescriptor(
        'TextSegments', True, TextSegmentsType, default_args={},
        docstring='The text segment basic information.')  # type: TextSegmentsType
    DataExtensions = _NITFElementDescriptor(
        'DataExtensions', True, DataExtensionsType, default_args={},
        docstring='The data extension basic information.')  # type: DataExtensionsType
    ReservedExtensions = _NITFElementDescriptor(
        'ReservedExtensions', True, ReservedExtensionsType, default_args={},
        docstring='The reserved extension basic information.')  # type: ReservedExtensionsType
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType
    ExtendedHeader = _NITFElementDescriptor(
        'ExtendedHeader', True, UserHeaderType, default_args={},
        docstring='Extended subheader - TRE list.')  # type: UserHeaderType

    def __init__(self, **kwargs):
        self._FHDR = 'NITF'
        self._FVER = '02.10'
        self._NUMX = 0
        super(NITFHeader, self).__init__(**kwargs)

    @property
    def FHDR(self):
        """
        str: File Profile Name. This field shall contain the character string uniquely denoting
        that the file is formatted using NITF. Always `NITF`.
        """

        return self._FHDR

    @FHDR.setter
    def FHDR(self, value):
        # static value "NITF", but the NITFElement extension usage requires phantom setting
        pass

    @property
    def FVER(self):
        """
        str: File Version. This field shall contain a BCS-A character string uniquely
        denoting the version. Always `02.10`.
        """

        return self._FVER

    @FVER.setter
    def FVER(self, value):
        # static value "02.10", but the NITFElement extension usage requires phantom setting
        pass

    @property
    def NUMX(self):
        """
        int: Reserved for future use. Always :code:`0`.
        """

        return self._NUMX

    @NUMX.setter
    def NUMX(self, value):
        # static value 0, but the NITFElement extension usage requires phantom setting
        pass

    @property
    def HL(self):
        """
        int: The length of this header object in bytes.
        """

        return self.get_bytes_length()

    @HL.setter
    def HL(self, value):
        # derived value 0, but the NITFElement extension usage requires phantom setting
        pass


#############
# NITF 2.0 version

class SymbolSegmentsType(_ItemArrayHeaders):
    """
    This holds the symbol subheader and item sizes.
    """

    _subhead_len = 4
    _item_len = 6


class LabelSegmentsType(_ItemArrayHeaders):
    """
    This holds the label subheader and item sizes.
    """

    _subhead_len = 4
    _item_len = 3


class NITFHeader0(NITFElement):
    """
    The main NITF file header for NITF version 2.0 - see standards document
    MIL-STD-2500A for more information.
    """

    _ordering = (
        'FHDR', 'FVER', 'CLEVEL', 'STYPE', 'OSTAID', 'FDT', 'FTITLE', 'Security',
        'FSCOP', 'FSCPYS', 'ENCRYP', 'ONAME', 'OPHONE', 'FL', 'HL',
        'ImageSegments', 'SymbolSegments', 'LabelSegments', 'TextSegments',
        'DataExtensions', 'ReservedExtensions', 'UserHeader', 'ExtendedHeader')
    _lengths = {
        'FHDR': 4, 'FVER': 5, 'CLEVEL': 2, 'STYPE': 4,
        'OSTAID': 10, 'FDT': 14, 'FTITLE': 80,
        'FSCOP': 5, 'FSCPYS': 5, 'ENCRYP': 1,
        'ONAME': 27, 'OPHONE': 18, 'FL': 12, 'HL': 6}
    CLEVEL = _IntegerDescriptor(
        'CLEVEL', True, 2, default_value=0,
        docstring='Complexity Level. This field shall contain the complexity level required to '
                  'interpret fully all components of the file. Valid entries are assigned in '
                  'accordance with complexity levels established in Table A-10.')  # type: int
    STYPE = _StringDescriptor(
        'STYPE', True, 4, default_value='BF01',
        docstring='Standard Type. Standard type or capability. A BCS-A character string `BF01` '
                  'which indicates that this file is formatted using ISO/IEC IS 12087-5. '
                  'NITF02.10 is intended to be registered as a profile of ISO/IEC IS 12087-5.')  # type: str
    OSTAID = _StringDescriptor(
        'OSTAID', True, 10, default_value='',
        docstring='Originating Station ID. This field shall contain the identification code or name of '
                  'the originating organization, system, station, or product. It shall not be '
                  'filled with BCS spaces')  # type: str
    FDT = _StringDescriptor(
        'FDT', True, 14, default_value='',
        docstring='File Date and Time. This field shall contain the time (UTC) of the files '
                  'origination in the format `YYYYMMDDhhmmss`.')  # type: str
    FTITLE = _StringDescriptor(
        'FTITLE', True, 80, default_value='',
        docstring='File Title. This field shall contain the title of the file.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags0, default_args={},
        docstring='The image security tags.')  # type: NITFSecurityTags0
    FSCOP = _StringDescriptor(
        'FSCOP', True, 5, default_value='    0',
        docstring='File Copy Number. This field shall contain the copy number of the file.')  # type: str
    FSCPYS = _StringDescriptor(
        'FSCPYS', True, 5, default_value='    0',
        docstring='File Number of Copies. This field shall contain the total number of '
                  'copies of the file.')  # type: str
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    ONAME = _StringDescriptor(
        'ONAME', True, 27, default_value='',
        docstring='Originator Name. This field shall contain a valid name for the operator '
                  'who originated the file.')  # type: str
    OPHONE = _StringDescriptor(
        'OPHONE', True, 18, default_value='',
        docstring='Originator Phone Number. This field shall contain a valid phone number '
                  'for the operator who originated the file.')  # type: str
    FL = _IntegerDescriptor(
        'FL', True, 12, docstring='The size in bytes of the entire file.')  # type: int
    ImageSegments = _NITFElementDescriptor(
        'ImageSegments', True, ImageSegmentsType, default_args={},
        docstring='The image segment basic information.')  # type: ImageSegmentsType
    SymbolSegments = _NITFElementDescriptor(
        'SymbolSegments', True, SymbolSegmentsType, default_args={},
        docstring='The symbols segment basic information.')  # type: SymbolSegmentsType
    LabelSegments = _NITFElementDescriptor(
        'LabelSegments', True, LabelSegmentsType, default_args={},
        docstring='The labels segment basic information.')  # type: LabelSegmentsType
    TextSegments = _NITFElementDescriptor(
        'TextSegments', True, TextSegmentsType, default_args={},
        docstring='The text segment basic information.')  # type: TextSegmentsType
    DataExtensions = _NITFElementDescriptor(
        'DataExtensions', True, DataExtensionsType, default_args={},
        docstring='The data extension basic information.')  # type: DataExtensionsType
    ReservedExtensions = _NITFElementDescriptor(
        'ReservedExtensions', True, ReservedExtensionsType, default_args={},
        docstring='The reserved extension basic information.')  # type: ReservedExtensionsType
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType
    ExtendedHeader = _NITFElementDescriptor(
        'ExtendedHeader', True, UserHeaderType, default_args={},
        docstring='Extended subheader - TRE list.')  # type: UserHeaderType

    def __init__(self, **kwargs):
        self._FHDR = 'NITF'
        self._FVER = '02.00'
        super(NITFHeader0, self).__init__(**kwargs)

    @property
    def FHDR(self):
        """
        str: File Profile Name. This field shall contain the character string uniquely denoting
        that the file is formatted using NITF. Always `NITF`.
        """

        return self._FHDR

    @FHDR.setter
    def FHDR(self, value):
        # static value "NITF", but the NITFElement extension usage requires phantom setting
        pass

    @property
    def FVER(self):
        """
        str: File Version. This field shall contain a BCS-A character string uniquely
        denoting the version, should generally be `02.00` or `01.10`.
        """

        return self._FVER

    @FVER.setter
    def FVER(self, value):
        if isinstance(value, bytes) and not isinstance(value, str):
            value = value.decode('utf-8')
        if not isinstance(value, str):
            raise TypeError('FVER is required to be a string')
        if len(value) != 5:
            raise ValueError('FVER must have length 5')
        if value not in ['02.00', '01.10']:
            logger.warning('Got unexpected version {}, and NITF parsing is likely to fail.'.format(value))
        self._FVER = value

    @property
    def HL(self):
        """
        int: The length of this header object in bytes.
        """

        return self.get_bytes_length()

    @HL.setter
    def HL(self, value):
        # derived value, but the NITFElement extension usage requires phantom setting
        pass




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class STDIDCType(TREElement):
    def __init__(self, value):
        super(STDIDCType, self).__init__()
        self.add_field('ACQUISITION_DATE', 's', 14, value)
        self.add_field('MISSION', 's', 14, value)
        self.add_field('PASS', 's', 2, value)
        self.add_field('OP_NUM', 's', 3, value)
        self.add_field('START_SEGMENT', 's', 2, value)
        self.add_field('REPRO_NUM', 's', 2, value)
        self.add_field('REPLAY_REGEN', 's', 3, value)
        self.add_field('BLANK_FILL', 's', 1, value)
        self.add_field('START_COLUMN', 's', 3, value)
        self.add_field('START_ROW', 's', 5, value)
        self.add_field('END_SEGMENT', 's', 2, value)
        self.add_field('END_COLUMN', 's', 3, value)
        self.add_field('END_ROW', 's', 5, value)
        self.add_field('COUNTRY', 's', 2, value)
        self.add_field('WAC', 's', 4, value)
        self.add_field('LOCATION', 's', 11, value)
        self.add_field('RESERV01', 's', 5, value)
        self.add_field('RESERV02', 's', 8, value)


class STDIDC(TREExtension):
    _tag_value = 'STDIDC'
    _data_type = STDIDCType



"""
This module provide utilities for converting from any complex format that we can
read to SICD or SIO format. The same conversion utility can be used to subset data.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Wade Schwartzkopf", "Thomas McCullough", "Valkyrie Systems Corporation")

import os
import logging
from typing import Union, List, Tuple, Callable, BinaryIO

import numpy

from sarpy.geometry.geocoords import ecf_to_geodetic
from sarpy.geometry.point_projection import image_to_ground_dem
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd import SICDWriter
from sarpy.io.complex.sio import SIOWriter
from sarpy.io.DEM.geotiff1deg import GeoTIFF1DegInterpolator
from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.base import check_for_openers
from sarpy.io.general.nitf import NITFReader
from sarpy.io.general.utils import is_file_like

logger = logging.getLogger(__name__)

###########
# Module variables
_writer_types = {'SICD': SICDWriter, 'SIO': SIOWriter}
_openers = []
_parsed_openers = False


def register_opener(open_func: Callable) -> None:
    """
    Provide a new opener.

    Parameters
    ----------
    open_func : callable
        This is required to be a function which takes a single argument (file name).
        This function should return a sarpy.io.complex.base.SICDTypeReader instance
        if the referenced file is viable for the underlying type, and None otherwise.

    Returns
    -------
    None
    """

    if not callable(open_func):
        raise TypeError('open_func must be a callable')
    if open_func not in _openers:
        _openers.append(open_func)


def parse_openers() -> None:
    """
    Automatically find the viable openers (i.e. :func:`is_a`) in the various modules.
    """

    global _parsed_openers
    if _parsed_openers:
        return
    _parsed_openers = True

    check_for_openers('sarpy.io.complex', register_opener)


def _define_final_attempt_openers() -> List[Callable]:
    """
    Gets the prioritized list of openers to attempt after regular openers.

    Returns
    -------
    List[Callable]
    """

    from sarpy.io.complex.other_nitf import final_attempt
    return [final_attempt, ]


def open_complex(file_name: Union[str, BinaryIO]) -> SICDTypeReader:
    """
    Given a file, try to find and return the appropriate reader object.

    Parameters
    ----------
    file_name : str|BinaryIO

    Returns
    -------
    SICDTypeReader

    Raises
    ------
    SarpyIOError
    """

    if (not is_file_like(file_name)) and (not os.path.exists(file_name)):
        raise SarpyIOError('File {} does not exist.'.format(file_name))
    # parse openers, if not already done
    parse_openers()
    # see if we can find a reader though trial and error
    for opener in _openers:
        reader = opener(file_name)
        if reader is not None:
            return reader
    # check the final attempt openers
    for opener in _define_final_attempt_openers():
        reader = opener(file_name)
        if reader is not None:
            return reader

    # If for loop completes, no matching file format was found.
    raise SarpyIOError('Unable to determine complex image format.')


class Converter(object):
    """
    This is a class for conversion (of a single frame) of one complex format to
    SICD or SIO format. Another use case is to create a (contiguous) subset of a
    given complex dataset. **This class is intended to be used as a context manager.**
    """

    __slots__ = ('_reader', '_file_name', '_writer', '_frame', '_row_limits', '_col_limits')

    def __init__(self, reader, output_directory, output_file=None, frame=None, row_limits=None, col_limits=None,
                 output_format='SICD', check_older_version=False, check_existence=True):
        """

        Parameters
        ----------
        reader : SICDTypeReader
            The base reader instance.
        output_directory : str
            The output directory. **This must exist.**
        output_file : None|str
            The output file name. If not provided, then `sicd.get_suggested_name(frame)`
            will be used.
        frame : None|int
            The frame (i.e. index into the reader's sicd collection) to convert.
            The default is 0.
        row_limits : None|Tuple[int, int]
           Row start/stop. Default is all.
        col_limits : None|Tuple[int, int]
           Column start/stop. Default is all.
        output_format : str
           The output file format to write, from {'SICD', 'SIO'}.  Default is SICD.
        check_older_version : bool
            Try to use a less recent version of SICD (1.1), for possible application compliance issues?
        check_existence : bool
            Should we check if the given file already exists, and raises an exception if so?
        """

        if isinstance(reader, SICDTypeReader):
            self._reader = reader
        else:
            raise ValueError(
                'reader is expected to be a Reader instance. Got {}'.format(type(reader)))

        if not (os.path.exists(output_directory) and os.path.isdir(output_directory)):
            raise SarpyIOError('output directory {} must exist.'.format(output_directory))
        if output_file is None:
            output_file = self._reader.get_sicds_as_tuple()[frame].get_suggested_name(frame+1)+'_SICD'
        output_path = os.path.join(output_directory, output_file)
        if check_existence and os.path.exists(output_path):
            raise SarpyIOError('The file {} already exists.'.format(output_path))

        # validate the output format and fetch the writer type
        if output_format is None:
            output_format = 'SICD'
        output_format = output_format.upper()
        if output_format not in ['SICD', 'SIO']:
            raise ValueError('Got unexpected output_format {}'.format(output_format))
        writer_type = _writer_types[output_format]

        # fetch the appropriate sicd instance
        sicds = self._reader.get_sicds_as_tuple()
        if frame is None:
            self._frame = 0
        else:
            self._frame = int(frame)
        if not (0 <= self._frame < len(sicds)):
            raise ValueError(
                'Got a frame {}, but it must be between 0 and {}'.format(frame, len(sicds)))
        this_sicd = sicds[self._frame]

        # validate row and column limits and update sicd structure, as necessary
        this_sicd, self._row_limits, self._col_limits = this_sicd.create_subset_structure(row_limits, col_limits)

        # set up our writer
        self._file_name = output_path
        self._writer = writer_type(
            output_path, this_sicd, check_older_version=check_older_version, check_existence=check_existence)

    def _get_rows_per_block(self, max_block_size):
        pixel_type = self._writer.sicd_meta.ImageData.PixelType
        cols = int(self._writer.sicd_meta.ImageData.NumCols)
        bytes_per_row = 8*cols
        if pixel_type == 'RE32F_IM32F':
            bytes_per_row = 8*cols
        elif pixel_type == 'RE16I_IM16I':
            bytes_per_row = 4*cols
        elif pixel_type == 'AMP8I_PHS8I':
            bytes_per_row = 2*cols
        return max(1, int(round(max_block_size/bytes_per_row)))

    @property
    def writer(self):  # type: () -> Union[SICDWriter, SIOWriter]
        """SICDWriter|SIOWriter: The writer instance."""
        return self._writer

    def write_data(self, max_block_size=None):
        r"""
        Assuming that the desired changes have been made to the writer instance
        nitf header tags, write the data.

        Parameters
        ----------
        max_block_size : None|int
            (nominal) maximum block size in bytes. Minimum value is :math:`2^{20} = 1~\text{MB}`.
            Default value is :math:`2^{26} = 64~\text{MB}`.

        Returns
        -------
        None
        """

        # validate max_block_size
        if max_block_size is None:
            max_block_size = 2**26
        else:
            max_block_size = int(max_block_size)
            if max_block_size < 2**20:
                max_block_size = 2**20

        # now, write the data
        rows_per_block = self._get_rows_per_block(max_block_size)
        block_start = self._row_limits[0]
        while block_start < self._row_limits[1]:
            block_end = min(block_start + rows_per_block, self._row_limits[1])
            data = self._reader[block_start:block_end, self._col_limits[0]:self._col_limits[1], self._frame, 'nosqueeze']
            self._writer.write_chip(data, start_indices=(block_start - self._row_limits[0], 0))
            logger.info('Done writing block {}-{} to file {}'.format(block_start, block_end, self._file_name))
            block_start = block_end
        self._writer.close()

    def __del__(self):
        if hasattr(self, '_writer'):
            self._writer.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type is None:
            self._writer.close()
        else:
            logger.error(
                'The {} file converter generated an exception during processing.\n\t'
                'The file {} may be only partially generated and corrupt.'.format(
                    self.__class__.__name__, self._file_name))
            # The exception will be reraised.
            # It's unclear how any exception could be caught.


def conversion_utility(
        input_file, output_directory, output_files=None, frames=None, output_format='SICD',
        row_limits=None, column_limits=None, max_block_size=None, check_older_version=False,
        preserve_nitf_information=False, check_existence=True,
        dem_filename_pattern=None, dem_type=None, geoid_file=None):
    """
    Copy SAR complex data to a file of the specified format.

    Parameters
    ----------
    input_file : str|SICDTypeReader
       Reader instance, or the name of file to convert.
    output_directory : str
        The output directory. **This must exist.**
    output_files : None|str|List[str]
       The name of the output file(s), or list of output files matching `frames`.
       If not provided, then `sicd.get_suggested_name(frame)` will be used.
    frames : None|int|list
       Set of frames to convert. Default is all.
    output_format : str
       The output file format to write, from {'SICD', 'SIO'}, optional.  Default is SICD.
    row_limits : None|Tuple[int, int]|List[Tuple[int, int]]
       Rows start/stop. Default is all.
    column_limits : None|Tuple[int, int]|List[Tuple[int, int]]
       Columns start/stop. Default is all.
    max_block_size : None|int
        (nominal) maximum block size in bytes. Passed through to the Converter class.
    check_older_version : bool
        Try to use a less recent version of SICD (1.1), for possible application compliance issues?
    preserve_nitf_information : bool
        Try to preserve NITF information? This only applies in the case that the file being read
        is actually a NITF file.
    check_existence : bool
        Check for the existence of any possibly overwritten file?
    dem_filename_pattern : str | None
        Optional string specifying a Digital Elevation Model (DEM) filename pattern.
        This is a format string that specifies a glob pattern that will
        uniquely specify a DEM file from the Lat/Lon of the SW corner of
        the DEM tile.  See the utils/convert_to_sicd help text for more details.
    dem_type : str | None
        Optional DEM type ('GeoTIFF', 'GeoTIFF:WGS84', 'GeoTIFF:EGM2008', etc.).
        This parameter is required when dem_filename_pattern is specified.  For 'GeoTIFF'
        DEM files, the reference surface can be either WGS84 or any of the geoid models.
        The reference surface is appended to the DEM type with a ':' separator.  If the
        reference surface is not specified, then EGM2008 is assumed.
    geoid_file : str | None
        Optional Geoid file which might be needed when dem_filename_pattern is specified.

    Returns
    -------
    None
    """

    def validate_lims(lims, typ):
        # type: (Union[None, tuple, list, numpy.ndarray], str) -> Tuple[Tuple[int, int], ...]
        def validate_entry(st, ed, shap, i_fr):
            if not ((0 <= st < shap[ind]) and (st < ed <= shap[ind])):
                raise ValueError('{}_limits is {}, and frame {} has shape {}'.format(typ, lims, i_fr, shap))

        ind = 0 if typ == 'row' else 1

        if lims is None:
            return tuple((0, shp[ind]) for shp in sizes)
        else:
            o_lims = numpy.array(lims, dtype=numpy.int64)
            t_lims = []
            if len(o_lims.shape) == 1:
                if o_lims.shape[0] != 2:
                    raise ValueError(
                        'row{}_limits must be of the form (<start>, <end>), got {}'.format(typ, lims))
                t_start = int(o_lims[0])
                t_end = int(o_lims[1])
                for i_frame, shp in zip(frames, sizes):
                    validate_entry(t_start, t_end, shp, i_frame)
                    t_lims.append((t_start, t_end))
            else:
                if o_lims.shape[0] != len(frames):
                    raise ValueError(
                        '{0:s}_limits must either be of the form (<start>, <end>)\n\t'
                        'applied to all frames, or a collection of such of the \n\t'
                        'same length as frames.\n\t'
                        'Got len({0:s}_limits) = {1:d} and len(frames) = {2:d}'.format(
                            typ, o_lims.shape[0], len(frames)))
                for entry, i_frame, shp in zip(o_lims, frames, sizes):
                    t_start = int(entry[0])
                    t_end = int(entry[1])
                    validate_entry(t_start, t_end, shp, i_frame)
                    t_lims.append((t_start, t_end))
            return tuple(t_lims)

    if isinstance(input_file, str):
        reader = open_complex(input_file)
    elif isinstance(input_file, SICDTypeReader):
        reader = input_file
    else:
        raise ValueError(
            'input_file is expected to be a file name or Reader instance.\n\t'
            'Got {}'.format(type(input_file)))

    if preserve_nitf_information and isinstance(reader, NITFReader):
        try:
            # noinspection PyUnresolvedReferences
            reader.populate_nitf_information_into_sicd()
        except AttributeError:
            logger.warning(
                'Reader class `{}` is missing populate_nitf_information_into_sicd '
                'method'.format(type(reader)))

    if not (os.path.exists(output_directory) and os.path.isdir(output_directory)):
        raise SarpyIOError('output directory {} must exist.'.format(output_directory))

    sicds = reader.get_sicds_as_tuple()
    sizes = reader.get_data_size_as_tuple()

    if dem_filename_pattern is not None:
        # Update the SICD metadata base on a projection of the SCP to a DEM.
        if dem_type.upper().startswith('GEOTIFF'):
            ref_surface = dem_type.upper().split(':')[-1] if ':' in dem_type else 'EGM2008'
            dem_interpolator = GeoTIFF1DegInterpolator(dem_filename_pattern,
                                                       ref_surface=ref_surface,
                                                       geoid_path=geoid_file)
        else:
            raise NotImplementedError(f'DEM type ({dem_type}) is not implemented.')

        for sicd in sicds:
            # project SCP and corner points to DEM
            scp = (sicd.ImageData.SCPPixel.Row, sicd.ImageData.SCPPixel.Col)
            frfc = (sicd.ImageData.FirstRow, sicd.ImageData.FirstCol)
            frlc = (sicd.ImageData.FirstRow, sicd.ImageData.FirstCol + sicd.ImageData.NumCols - 1)
            lrfc = (sicd.ImageData.FirstRow + sicd.ImageData.NumRows - 1, sicd.ImageData.FirstCol)
            lrlc = (sicd.ImageData.FirstRow + sicd.ImageData.NumRows - 1,
                    sicd.ImageData.FirstCol + sicd.ImageData.NumCols - 1)

            img_points = [scp, frfc, frlc, lrfc, lrlc]
            ecf_points = image_to_ground_dem(img_points, sicd,
                                             block_size=None,
                                             dem_interpolator=dem_interpolator,
                                             pad_value=0.2,
                                             vertical_step_size=1.0,
                                             use_structure_coa=True,
                                             )

            llh_points = ecf_to_geodetic(ecf_points)

            sicd.GeoData.SCP.LLH.Lat = llh_points[0][0]
            sicd.GeoData.SCP.LLH.Lon = llh_points[0][1]
            sicd.GeoData.SCP.LLH.HAE = llh_points[0][2]

            sicd.GeoData.SCP.ECF.X = ecf_points[0][0]
            sicd.GeoData.SCP.ECF.Y = ecf_points[0][1]
            sicd.GeoData.SCP.ECF.Z = ecf_points[0][2]

            for i in range(4):
                index = int(sicd.GeoData.ImageCorners[i].index.split(':')[0])    # index = 1|2|3|4, SCP is index 0
                sicd.GeoData.ImageCorners[i].Lat = llh_points[index][0]          # FRFC is index 1, FRLC is index 2,
                sicd.GeoData.ImageCorners[i].Lon = llh_points[index][1]          # LRFC is index 3, LRLC is index 4

            # Recompute SCPCOA using the updated SCP
            sicd.SCPCOA.rederive(sicd.Grid, sicd.Position, sicd.GeoData)

    # check that frames is valid
    if frames is None:
        frames = tuple(range(len(sicds)))
    if isinstance(frames, int):
        frames = (frames, )
    if not isinstance(frames, tuple):
        frames = tuple(int(entry) for entry in frames)
    if len(frames) == 0:
        raise ValueError('The list of frames is empty.')
    o_frames = []
    for frame in frames:
        index = int(frame)
        if not (0 <= index < len(sicds)):
            raise ValueError(
                'Got a frames entry {}, but it must be between 0 and {}'.format(index, len(sicds)))
        o_frames.append(index)
    frames = tuple(o_frames)

    # assign SUGGESTED_NAME to each sicd
    for frame in frames:
        sicd = sicds[frame]
        suggested_name = sicd.get_suggested_name(frame+1)+'_SICD'
        if suggested_name is None and sicd.CollectionInfo.CoreName is not None:
            suggested_name = sicd.CollectionInfo.CoreName+'{}_SICD'.format(frame)
        if suggested_name is None:
            suggested_name = 'Unknown{}_SICD'.format(frame)
        sicd.NITF['SUGGESTED_NAME'] = suggested_name
    # construct output_files list
    if output_files is None:
        output_files = [sicds[frame].NITF['SUGGESTED_NAME']+'.nitf' for frame in frames]
    elif isinstance(output_files, str):
        if len(sicds) == 1:
            output_files = [output_files, ]
        else:
            digits = int(numpy.ceil(numpy.log10(len(sicds))))
            frm_str = '{0:s}-{1:0' + str(digits) + 'd}{2:s}'
            fstem, fext = os.path.splitext(output_files)
            o_files = []
            for index in frames:
                o_files.append(frm_str.format(fstem, index, fext))
            output_files = tuple(o_files)

    if len(output_files) != len(frames):
        raise ValueError('The lengths of frames and output_files must match.')
    if len(set(output_files)) != len(output_files):
        raise ValueError(
            'Entries in output_files (possibly constructed) must be unique,\n\t'
            'got {} for frames {}'.format(output_files, frames))

    # construct validated row/column_limits
    row_limits = validate_lims(row_limits, 'row')
    column_limits = validate_lims(column_limits, 'column')

    for o_file, frame, row_lims, col_lims in zip(output_files, frames, row_limits, column_limits):
        logger.info('Converting frame {} from file {} to file {}'.format(frame, input_file, o_file))
        with Converter(
                reader, output_directory, output_file=o_file, frame=frame,
                row_limits=row_lims, col_limits=col_lims, output_format=output_format,
                check_older_version=check_older_version,
                check_existence=check_existence) as converter:
            converter.write_data(max_block_size=max_block_size)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PT(TREElement):
    def __init__(self, value):
        super(PT, self).__init__()
        self.add_field('LON', 's', 15, value)
        self.add_field('LAT', 's', 15, value)


class ACPO(TREElement):
    def __init__(self, value):
        super(ACPO, self).__init__()
        self.add_field('UNIAAH', 's', 3, value)
        self.add_field('AAH', 's', 5, value)
        self.add_field('UNIAAV', 's', 3, value)
        self.add_field('AAV', 's', 5, value)
        self.add_field('UNIAPH', 's', 3, value)
        self.add_field('APH', 's', 5, value)
        self.add_field('UNIAPV', 's', 3, value)
        self.add_field('APV', 's', 5, value)
        self.add_field('NUMPTS', 'd', 3, value)
        self.add_loop('PTs', self.NUMPTS, PT, value)


class ACCPOBType(TREElement):
    def __init__(self, value):
        super(ACCPOBType, self).__init__()
        self.add_field('NUMACPO', 'd', 2, value)
        self.add_loop('ACPOs', self.NUMACPO, ACPO, value)


class ACCPOB(TREExtension):
    _tag_value = 'ACCPOB'
    _data_type = ACCPOBType



"""
This module provides common functions for creating kmz products.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Valkyrie Systems Corporation"

import logging

import matplotlib.pyplot as plt
import numpy as np
import shapely.geometry as shg

from sarpy.geometry.geocoords import ecf_to_geodetic

WGS84_SEMIMAJOR = 6378137.0
WGS84_SEMIMINOR = 6356752.314245

logger = logging.getLogger(__name__)


def ecef_to_kml_coord(ecf_points):
    """Convert a list of ECEF points to a list of KML coordinates"""
    # TODO apply geoid.  KML expects MSL
    llh_points = ecf_to_geodetic(ecf_points)
    return [f"{lon},{lat},{alt}" for (lat, lon, alt) in llh_points]


def acf_to_ecef(eb_dcx, eb_dcy, uacx, uacy):
    """Convert an ACF pointing vector into an ECEF pointing vector"""
    uacz = np.cross(uacx, uacy)
    eb_dcz = np.sqrt(1 - eb_dcx**2 - eb_dcy**2)
    eb = np.stack((eb_dcx, eb_dcy, eb_dcz)).T

    eb_pointing = eb[:, 0, np.newaxis] * uacx
    eb_pointing += eb[:, 1, np.newaxis] * uacy
    eb_pointing += eb[:, 2, np.newaxis] * uacz

    return eb_pointing


def add_polygon_style(
    kmz_document, name, *, bbggrr, low_aa, low_width, high_aa, high_width, outline="1"
):
    """Add a polygon style to a KML document"""
    opaque = "ff"
    kmz_document.add_style(
        name + "_high",
        line_style={"color": opaque + bbggrr, "width": high_width},
        poly_style={"color": high_aa + bbggrr, "outline": str(outline)},
    )
    kmz_document.add_style(
        name + "_low",
        line_style={"color": opaque + bbggrr, "width": low_width},
        poly_style={"color": low_aa + bbggrr, "outline": str(outline)},
    )
    kmz_document.add_style_map(name, name + "_high", name + "_low")


def add_antenna_styles(kmz_document):
    """Add KML styles used for rendering antenna information"""
    add_polygon_style(
        kmz_document,
        "mechanical_boresight",
        bbggrr="a00000",
        low_aa="70",
        low_width="2.0",
        high_aa="a0",
        high_width="3.5",
        outline="0",
    )

    add_polygon_style(
        kmz_document,
        "electrical_boresight",
        bbggrr="a0a050",
        low_aa="70",
        low_width="2.0",
        high_aa="a0",
        high_width="3.5",
        outline="0",
    )

    add_polygon_style(
        kmz_document,
        "rcv_beam_footprint",
        bbggrr="ffaa55",
        low_aa="70",
        low_width="1.0",
        high_aa="a0",
        high_width="1.5",
    )

    add_polygon_style(
        kmz_document,
        "tx_beam_footprint",
        bbggrr="0000ff",
        low_aa="70",
        low_width="1.0",
        high_aa="a0",
        high_width="1.5",
    )

    add_polygon_style(
        kmz_document,
        "twoway_beam_footprint",
        bbggrr="885588",
        low_aa="70",
        low_width="1.0",
        high_aa="a0",
        high_width="1.5",
    )


def ray_ellipsoid_intersection(
    semiaxis_a, semiaxis_b, semiaxis_c, pos_x, pos_y, pos_z, dir_x, dir_y, dir_z
):
    """
    Compute the intersection of a Ray with an Ellipsoid

    Parameters
    ----------
    semiaxis_a, semiaxis_b, semiaxis_c: float
        Semiaxes of the ellipsoid
    pos_x, pos_y, pos_z: float
        Ray origin
    dir_x, dir_y, dir_z: float
        Ray direction

    Returns
    -------
    intersection_x, intersection_y, intersection_z: float
        Point of intersection
    """
    # Based on https://stephenhartzell.medium.com/satellite-line-of-sight-intersection-with-earth-d786b4a6a9b6
    # >>> pos_x, pos_y, pos_z = symbols('pos_x pos_y pos_z')
    # >>> dir_x, dir_y, dir_z = symbols('dir_x dir_y dir_z')
    # >>> semiaxis_a, semiaxis_b, semiaxis_c = symbols('semiaxis_a semiaxis_b semiaxis_c')
    # >>> distance = symbols('distance')

    # >>> solutions = solve((pos_x + distance*dir_x)**2/semiaxis_a**2 + (pos_y + distance*dir_y)**2/semiaxis_b**2 + (pos_z + distance*dir_z)**2/semiaxis_c**2 - 1, distance)
    # >>> print(solutions[0])

    # black formatted
    distance = (
        -dir_x * pos_x * semiaxis_b**2 * semiaxis_c**2
        - dir_y * pos_y * semiaxis_a**2 * semiaxis_c**2
        - dir_z * pos_z * semiaxis_a**2 * semiaxis_b**2
        - semiaxis_a
        * semiaxis_b
        * semiaxis_c
        * np.sqrt(
            -(dir_x**2) * pos_y**2 * semiaxis_c**2
            - dir_x**2 * pos_z**2 * semiaxis_b**2
            + dir_x**2 * semiaxis_b**2 * semiaxis_c**2
            + 2 * dir_x * dir_y * pos_x * pos_y * semiaxis_c**2
            + 2 * dir_x * dir_z * pos_x * pos_z * semiaxis_b**2
            - dir_y**2 * pos_x**2 * semiaxis_c**2
            - dir_y**2 * pos_z**2 * semiaxis_a**2
            + dir_y**2 * semiaxis_a**2 * semiaxis_c**2
            + 2 * dir_y * dir_z * pos_y * pos_z * semiaxis_a**2
            - dir_z**2 * pos_x**2 * semiaxis_b**2
            - dir_z**2 * pos_y**2 * semiaxis_a**2
            + dir_z**2 * semiaxis_a**2 * semiaxis_b**2
        )
    ) / (
        dir_x**2 * semiaxis_b**2 * semiaxis_c**2
        + dir_y**2 * semiaxis_a**2 * semiaxis_c**2
        + dir_z**2 * semiaxis_a**2 * semiaxis_b**2
    )

    if np.isnan(distance):
        raise ValueError("Ray does not intersect ellipsoid")

    if distance < 0:
        raise ValueError("Ray points away from ellipsoid")

    return (
        pos_x + distance * dir_x,
        pos_y + distance * dir_y,
        pos_z + distance * dir_z,
    )


def ray_intersect_earth(position, direction):
    """Intersect an ECEF ray with the earth

    Parameters
    ----------
    position: list-like
        Ray origin in ECEF
    direction: list-like
        Ray direction in ECEF

    Returns
    -------
    intersection_x, intersection_y, intersection_z: float
        Point of intersection in ECEF

    """
    point_ecf = ray_ellipsoid_intersection(
        WGS84_SEMIMAJOR,
        WGS84_SEMIMAJOR,
        WGS84_SEMIMINOR,
        position[0],
        position[1],
        position[2],
        direction[0],
        direction[1],
        direction[2],
    )
    return point_ecf


def add_los_polygon(kmz_doc, parent_placemark, apc_coords, ground_coords):
    """Add a polygon connecting a path above the earth to a path on the earth"""
    # complex 3d polygons don't always render nicely.  So, we'll manually triangluate it.
    mg = kmz_doc.add_multi_geometry(par=parent_placemark)
    # Highlight the starting point
    kmz_doc.add_line_string(
        coords=" ".join([apc_coords[0], ground_coords[0]]),
        par=mg,
        altitudeMode="absolute",
    )
    for idx in range(len(apc_coords) - 1):
        coords = [
            apc_coords[idx],
            ground_coords[idx],
            apc_coords[idx + 1],
            apc_coords[idx],
        ]
        kmz_doc.add_polygon(" ".join(coords), par=mg, altitudeMode="absolute")
        coords = [
            ground_coords[idx],
            ground_coords[idx + 1],
            apc_coords[idx + 1],
            ground_coords[idx],
        ]
        kmz_doc.add_polygon(" ".join(coords), par=mg, altitudeMode="absolute")


def make_beam_footprints(
    aiming_metadata,
    labelled_indices,
    array_gain_poly,
    element_gain_poly,
    contour_level=-3,
):
    """Attempt to make beam footprints for the labelled slowtimes"""

    def _find_central_contour(max_dc, eb_dcx, eb_dcy):
        Ns = 201
        dcs = np.linspace(-max_dc, max_dc, Ns)
        dcx, dcy = np.meshgrid(dcs, dcs, indexing="ij")
        gain_pattern = array_gain_poly(dcx, dcy) + element_gain_poly(dcx + eb_dcx, dcy + eb_dcy)
        contour_sets = plt.contour(dcx, dcy, gain_pattern, levels=[contour_level])
        plt.close()  # close the figure created by contour

        # We don't know the validity range of the gain polynomials and may have gone
        # outside, resulting in multiple contours

        # Only keep contours that form a closed shape
        try:
            paths = contour_sets.get_paths()
        except AttributeError:
            # matplotlib deprecated collections attribute in 3.8
            paths = contour_sets.collections[0].get_paths()

        polygons = [
            polygon
            for path in paths
            for polygon in path.to_polygons(closed_only=False)
            if np.array_equal(polygon[0], polygon[-1]) and len(polygon) > 1  # only consider closed polygons
        ]
        if polygons:
            # Keep contour closest to center
            return min(
                polygons, key=lambda vertices: np.linalg.norm(np.mean(vertices, axis=0))
            )
        return None


    result = {}
    for name, pvp_index in labelled_indices.items():
        try:
            eb_dcx = aiming_metadata["raw"]["eb_dcx"][pvp_index]
            eb_dcy = aiming_metadata["raw"]["eb_dcy"][pvp_index]
            uacx = aiming_metadata["raw"]["uacx"][pvp_index]
            uacy = aiming_metadata["raw"]["uacy"][pvp_index]
            time = aiming_metadata["raw"]["times"][pvp_index]
            apc_pos = aiming_metadata["raw"]["positions"][pvp_index]

            contour = _find_central_contour(1, eb_dcx, eb_dcy)
            for n in range(10):
                next_contour = _find_central_contour(8**(-n-1), eb_dcx, eb_dcy)
                if next_contour is None and contour is not None:
                    break
                if contour is not None and next_contour is not None:
                    cpoly = shg.Polygon(contour)
                    ncpoly = shg.Polygon(next_contour)
                    if np.isclose(cpoly.intersection(ncpoly).area, ncpoly.area, atol=0, rtol=0.05):
                        # close enough
                        contour = next_contour
                        break
                contour = next_contour

            if contour is None:
                raise ValueError("unable to find contour")

            delta_dcx = contour[:, 0]
            delta_dcy = contour[:, 1]

            contour_pointing = acf_to_ecef(
                delta_dcx + eb_dcx,
                delta_dcy + eb_dcy,
                uacx,
                uacy,
            )

            contour_earth_ecf = []
            for along in contour_pointing:
                contour_earth_ecf.append(ray_intersect_earth(apc_pos, along))
            result[name] = {"time": time, "contour": contour_earth_ecf}
        except Exception as exc:
            logger.warning(
                f"Exception while calculating {name} beam footprint of {aiming_metadata['antpat_id']}"
            )
            logger.warning(exc, exc_info=True)

    return result
